/*
 * Licensed to UbiCollab.org under one or more contributor
 * license agreements.  See the NOTICE file distributed
 * with this work for additional information regarding
 * copyright ownership. UbiCollab.org licenses this file
 * to you under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package no.ntnu.android.toolkit.appstore;

import no.ntnu.android.toolkit.model.*;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class AppServerDataStub {

    private static AppServerDataStub dataStub;
    private List<SavedToolkitApplication> savedToolkitApplications;
    private static IdInterface connector = new IdInterface() {
        @Override
        public int getId() {
            return 0;
        }
    };


    private AppServerDataStub() {
        savedToolkitApplications = new ArrayList<SavedToolkitApplication>();
        savedToolkitApplications.add(getCharlieAccessControl());
        savedToolkitApplications.add(getBravoJingles());
        savedToolkitApplications.add(getAlfaBravoMario());
    }

    public static AppServerDataStub getInstance() {
        if (dataStub == null) {
            dataStub = new AppServerDataStub();
        }
        return dataStub;
    }
    //----------------------------------------------------------------------//
    //                      Queries                                         //
    //----------------------------------------------------------------------//

    public boolean save(SavedToolkitApplication application) {
        return savedToolkitApplications.add(application);
    }

    public List<String> getAllBoardTypeNames() {
        List<String> names = new ArrayList<String>();
        for(SavedToolkitApplication app : getInstance().getSavedToolkitApplications()) {
            for(Board board : app.getBoards() ) {
                if(!names.contains(board.getName())) {
                    names.add(board.getHardwareSetup().getName());
                }
            }
        }
        return names;
    }

    public List<SavedToolkitApplication> getSavedToolkitApplications() {
        return savedToolkitApplications;
    }

    public List<SavedToolkitApplication> getApplicationsIncludingName(List<String> names) {
        List<SavedToolkitApplication> applications = new ArrayList<SavedToolkitApplication>();
        for(SavedToolkitApplication app : getInstance().getSavedToolkitApplications()) {
            for(Board board : app.getBoards() ) {
                if(names.contains(board.getHardwareSetup().getName())) {
                    if(!applications.contains(app)) {
                        applications.add(app);
                    }
                }
            }
        }
        return applications;
    }

    public List<SavedToolkitApplication> getAppsWithBoardsIncludingNames(List<String> names) {
        List<SavedToolkitApplication> applications = new ArrayList<SavedToolkitApplication>();
        for(SavedToolkitApplication app : getInstance().getSavedToolkitApplications()) {
            if (appOnlyIncludesBoardsWithNames(app, names)) {
                applications.add(app);
            }
        }
        return applications;
    }

    private boolean appOnlyIncludesBoardsWithNames(SavedToolkitApplication app, List<String> names) {
        for (Board board : app.getBoards()) {
            if (!names.contains(board.getHardwareSetup().getName())) {
                return false;
            }
        }
        return true;
    }

    //----------------------------------------------------------------------//
    //                      List "entries"/instances                        //
    //----------------------------------------------------------------------//


    private static SavedToolkitApplication getBravoJingles() {
        SavedToolkitApplication savedToolkitApplication = new SavedToolkitApplication("Jingle bells alarm", false);
        HardwareSetup hardwareSetup = getHardwareSetup("Bravo","This board comes with a tilt " +
                "sensor, green led, RFID reader, a speaker in addition to a wireless network device.");
        Device greenLed = getDeviceLed("Green", "11");
        hardwareSetup.add(greenLed);
        Device speaker = getDeviceSpeaker("Speaker", "6");
        hardwareSetup.add(speaker);
        Device wifi = getDeviceWifi("Message", "9", "10");
        hardwareSetup.add(wifi);
        Device tilt = getDeviceTiltSensor("Tilt", "8");
        hardwareSetup.add(tilt);
        Device rfidreader = getDeviceRfidReader("7");
        hardwareSetup.add(rfidreader);

        hardwareSetup.addRFIDGroup(getRFIDGroupPalfi());
        hardwareSetup.addRFIDGroup(getRFIDGroupDaniel());
        hardwareSetup.addRFIDGroup(getRFIDGroupHaakon());
        Board board = getBoard(hardwareSetup);

        //"Software: Rules"
        Condition condition = getCondition(getDeviceFunctionTilt(), tilt, board);
        Action action = getAction(getDeviceFunctionSpeakerJingle(), speaker, board);
        Rule rule = createRule(board, condition, action);
        board.addRule(rule);
        List<ConfigurationPoint> configurationPoints = new ArrayList<ConfigurationPoint>();

        configurationPoints.add(new ConfigurationPoint(condition.getId(), 100, 100, IdGenerator.getInstance().getNext(connector)));
        configurationPoints.add(new ConfigurationPoint(action.getId(), 400, 400, IdGenerator.getInstance().getNext(connector)));
        savedToolkitApplication.setConnectorConfigurationPoints(
                new ArrayList<List<ConfigurationPoint>>(Arrays.asList(configurationPoints)));
        savedToolkitApplication.addBoard(board);
        savedToolkitApplication.setDescription("Tilt board and play jingle bells");

        return savedToolkitApplication;
    }

    private static SavedToolkitApplication getAlfaBravoMario() {
        SavedToolkitApplication savedToolkitApplication = new SavedToolkitApplication("Super Mario over there", false);
        HardwareSetup hardwareSetup = getHardwareSetup("Bravo","This board comes with a tilt sensor, " +
                "green led, RFID reader, a speaker in addition to a wireless network device.");
        Device greenLed = getDeviceLed("Green", "11");
        hardwareSetup.add(greenLed);
        Device speaker = getDeviceSpeaker("Speaker", "6");
        hardwareSetup.add(speaker);
        Device wifi = getDeviceWifi("Message", "9", "10");
        hardwareSetup.add(wifi);
        Device tilt = getDeviceTiltSensor("Tilt","8");
        hardwareSetup.add(tilt);
        Device rfidreader = getDeviceRfidReader("7");
        hardwareSetup.add(rfidreader);
        hardwareSetup.addRFIDGroup(getRFIDGroupPalfi());
        hardwareSetup.addRFIDGroup(getRFIDGroupDaniel());
        hardwareSetup.addRFIDGroup(getRFIDGroupHaakon());
        Board board = getBoard(hardwareSetup);

        HardwareSetup hardwareSetup2 = getHardwareSetup("Alfa","Has wireless network, 2 buttons, " +
                "1 RFID reader and 1 yellow led. Perfect for RFID apps.");
        Device greenLed2 = getDeviceLed("Green", "11");
        hardwareSetup2.add(greenLed2);
        Device wifi2 = getDeviceWifi("Message", "6", "7");
        hardwareSetup2.add(wifi2);
        Device button = getDeviceButton("Button", "10");
        hardwareSetup2.add(button);
        Device rfidreader2 = getDeviceRfidReader("6");
        hardwareSetup2.add(rfidreader2);
        hardwareSetup2.addRFIDGroup(getRFIDGroupPalfi());
        hardwareSetup2.addRFIDGroup(getRFIDGroupDaniel());
        hardwareSetup2.addRFIDGroup(getRFIDGroupHaakon());
        Board board2 = getBoard(hardwareSetup2);

        //"Software: Rules"
        Condition condition = getCondition(getDeviceFunctionButtonPressed(), button, board2);
        AutomaticSendMessage action = getAutomaticSendMessage(wifi2, board2);
        action.setDestinationBoard(board);
        Rule rule = createRule(board2, condition, action);
        board2.addRule(rule);
        AutomaticReceiveMessage condition2 = getAutomaticReceiveMessage(wifi, board);
        Action action2 = getAction(getDeviceFunctionSpeakerMario(), speaker, board);
        Rule rule2 = createRule(board, condition2, action2);
        board.addRule(rule2);
        List<ConfigurationPoint> configurationPoints = new ArrayList<ConfigurationPoint>();

        configurationPoints.add(new ConfigurationPoint(condition.getId(), 100, 100, IdGenerator.getInstance().getNext(connector)));
        configurationPoints.add(new ConfigurationPoint(action2.getId(), 700, 100, IdGenerator.getInstance().getNext(connector)));
        savedToolkitApplication.setConnectorConfigurationPoints(
                new ArrayList<List<ConfigurationPoint>>(Arrays.asList(configurationPoints)));
        savedToolkitApplication.addBoard(board);
        savedToolkitApplication.addBoard(board2);
        savedToolkitApplication.setDescription("Press button on 'Alfa' and play Super Mario on a 'Bravo'");

        return savedToolkitApplication;
    }

    private static SavedToolkitApplication getCharlieAccessControl() {
        SavedToolkitApplication savedToolkitApplication = new SavedToolkitApplication("Access control", false);
        HardwareSetup hardwareSetup = getHardwareSetup("Charlie","RFID reader, 1 button and 3 LEDs. Perfect for a start app.");
        Device greenLed = getDeviceLed("Green", "9");
        hardwareSetup.add(greenLed);
        Device redLed = getDeviceLed("Red", "8");
        hardwareSetup.add(redLed);
        Device yellowLed = getDeviceLed("Yellow", "11");
        hardwareSetup.add(yellowLed);
        Device button = getDeviceButton("Button", "10");
        hardwareSetup.add(button);
        Device rfid = getDeviceRfidReader("7");
        hardwareSetup.add(rfid);
        Board board = getBoard(hardwareSetup);

        hardwareSetup.addRFIDGroup(getRFIDGroupDaniel());
        hardwareSetup.addRFIDGroup(getRFIDGroupHaakon());
        hardwareSetup.addRFIDGroup(getRFIDGroupPalfi());

        //GUI placement points
        List<List<ConfigurationPoint>> listList = new ArrayList<List<ConfigurationPoint>>();

        //Rule 2: Button -> Green
        Condition buttonClicked = getCondition(getDeviceFunctionButtonPressed(), button, board);
        Action greenLedFlash = getAction(getDeviceFunctionLedFlash(1000), greenLed, board);
        Rule rule2 = createRule(board, buttonClicked, greenLedFlash);
        board.addRule(rule2);
        List<ConfigurationPoint> configurationPoints2 = new ArrayList<ConfigurationPoint>();
        configurationPoints2.add(
                new ConfigurationPoint(buttonClicked.getId(), 50, 350, IdGenerator.getInstance().getNext(connector)));
        ConfigurationPoint greenLedFlashConfigurationPoint =
                new ConfigurationPoint(greenLedFlash.getId(), 900, 350, IdGenerator.getInstance().getNext(connector));
        configurationPoints2.add(greenLedFlashConfigurationPoint);
        listList.add(configurationPoints2);

        //Rule 1: RFID access OK -> RED, Green (created above)
        Condition rfidAccess = getCondition(getDeviceFunctionRfidRead(getRFIDGroupDaniel()), rfid, board);
        Action redLedOff = getAction(getDeviceFunctionLed(false), redLed, board);
        Rule rule1 = createRule(board, rfidAccess, redLedOff);
        rule1.addConfiguration(greenLedFlash);
        board.addRule(rule1);
        List<ConfigurationPoint> configurationPoints1 = new ArrayList<ConfigurationPoint>();
        configurationPoints1.add(new ConfigurationPoint(rfidAccess.getId(), 50, 50, IdGenerator.getInstance().getNext(connector)));
        configurationPoints1.add(new ConfigurationPoint(redLedOff.getId(), 700, 50, IdGenerator.getInstance().getNext(connector)));
        configurationPoints1.add(greenLedFlashConfigurationPoint);
        listList.add(configurationPoints1);

        //Rule 3: RFID access Denied -> RED, Green
        Condition rfidDenied = getCondition(getDeviceFunctionRfidRead(getRFIDGroupHaakon()), rfid, board);
        Action redOnFlash = getAction(getDeviceFunctionLedFlash(1000), redLed, board);
        Action greenLedOff = getAction(getDeviceFunctionLed(false), greenLed, board);

        Rule rule3 = createRule(board, rfidDenied, redOnFlash);
        rule3.addConfiguration(greenLedOff);
        board.addRule(rule3);
        List<ConfigurationPoint> configurationPoints3 = new ArrayList<ConfigurationPoint>();
        configurationPoints3.add(new ConfigurationPoint(rfidDenied.getId(), 50, 650, IdGenerator.getInstance().getNext(connector)));
        configurationPoints3.add(new ConfigurationPoint(redOnFlash.getId(), 800, 600, IdGenerator.getInstance().getNext(connector)));
        configurationPoints3.add(new ConfigurationPoint(greenLedOff.getId(), 650, 800, IdGenerator.getInstance().getNext(connector)));
        listList.add(configurationPoints3);

        savedToolkitApplication.setConnectorConfigurationPoints(listList);
        savedToolkitApplication.addBoard(board);
        savedToolkitApplication.setDescription("Set and adjust access rights based on RFID tags. Use the button for grant access manually");
        return savedToolkitApplication;
    }

    private static RFIDGroup getRFIDGroupDaniel() {
        RFIDGroup rfidGroup = new RFIDGroup();
        rfidGroup.setName("Daniel SINTEF");
        rfidGroup.addTag("6F006909B0BF");
        rfidGroup.addTag("010B98B2E0C0");
        return rfidGroup;
    }

    private static RFIDGroup getRFIDGroupHaakon() {
        RFIDGroup rfidGroup = new RFIDGroup();
        rfidGroup.setName("Haakon SINTEF");
        rfidGroup.addTag("210047517740");
        rfidGroup.addTag("010B9900D546");
        return rfidGroup;
    }

    private static RFIDGroup getRFIDGroupPalfi() {
        RFIDGroup rfidGroup = new RFIDGroup();
        rfidGroup.setName("Anders SINTEF");
        rfidGroup.addTag("010B98A64C78");
        rfidGroup.addTag("21004268000B");
        return rfidGroup;
    }

    private static RFIDGroup getRFIDGroupBlack() {
        RFIDGroup rfidGroup = new RFIDGroup();
        rfidGroup.setName("Black tag");
        rfidGroup.addTag("010249D270E8");
        rfidGroup.addTag("0101A84F9C7B");
        return rfidGroup;
    }

    private static RFIDGroup getRFIDGroupWhite() {
        RFIDGroup rfidGroup = new RFIDGroup();
        rfidGroup.setName("White tag");
        rfidGroup.addTag("5100FDAE191B");
        rfidGroup.addTag("35021EA89B1A");
        return rfidGroup;
    }

    private static Board getBoard(HardwareSetup hardwareSetup) {
        return new Board(hardwareSetup);
    }

    private static HardwareSetup getHardwareSetup(String name, String description) {
        HardwareSetup hardwareSetup = new HardwareSetup();
        hardwareSetup.setName(name);
        hardwareSetup.setDescription(description);
        return hardwareSetup;
    }

    private static Device getDeviceRfidReader(String pinnumber) {
        Device device = new Device();
        device.setType("rfidreader");
        Pin pin = new Pin(pinnumber, Pin.PinMode.NONE);
        device.addPin(pin);
        return device;
    }

    private static Device getDeviceTiltSensor(String text, String pinnumber) {
        Device device = new Device();
        device.setType("tiltsensor");
        device.setText(text);
        Pin pin = new Pin(pinnumber, Pin.PinMode.INPUT);
        device.addPin(pin);
        return device;
    }

    private static Device getDeviceWifi(String text, String rx, String tx) {
        Device device = new Device();
        device.setType("wifi");
        device.setText(text);
        Pin pin = new Pin(rx, Pin.PinMode.NONE);
        device.addPin(pin);
        pin = new Pin(tx, Pin.PinMode.NONE);
        device.addPin(pin);
        return device;
    }

    private static Device getDeviceLed(String text, String pinnumber) {
        Device device = new Device();
        device.setType("LED");
        device.setText(text);
        Pin pin = new Pin(pinnumber, Pin.PinMode.OUTPUT);
        device.addPin(pin);
        return device;
    }

    private static Device getDeviceSpeaker(String text, String pinnumber) {
        Device device = new Device();
        device.setType("speaker");
        device.setText(text);
        Pin pin = new Pin(pinnumber, Pin.PinMode.OUTPUT);
        device.addPin(pin);
        return device;
    }

    private static Device getDeviceButton(String text, String pinnumber) {
        Device device = new Device();
        device.setType("Button");
        device.setText(text);
        Pin pin = new Pin(pinnumber, Pin.PinMode.INPUT);
        device.addPin(pin);
        return device;
    }

    private static Rule createRule(Board board, Condition condition, Action action) {
        Rule rule = new Rule(board);
        rule.addConfiguration(condition);
        rule.addConfiguration(action);
        return rule;
    }

    private static DeviceFunction getDeviceFunctionMessageReceived(String message) {
        DeviceFunction deviceFunction = new DeviceFunction("receiveMessage", DeviceFunction.Type.ACTUATOR);
        deviceFunction.setGuiName("Receive message");
        Parameter parameter = new Parameter("TextField", "String");
        parameter.setValueAsString(message);
        deviceFunction.addParameter(parameter);
        return deviceFunction;
    }

    private static DeviceFunction getDeviceFunctionSendMessage(String ip, String port, String message, String guiname) {
        DeviceFunction deviceFunction = new DeviceFunction("sendMessage", DeviceFunction.Type.SENSOR);
        deviceFunction.setGuiName(guiname);
        Parameter parameter = new Parameter("TextField", "String");
        parameter.setValueAsString(ip);
        deviceFunction.addParameter(parameter);
        parameter = new Parameter("TextField", "Integer");
        parameter.setValueAsString(port);
        deviceFunction.addParameter(parameter);
        parameter = new Parameter("TextField", "String");
        parameter.setValueAsString(message);
        deviceFunction.addParameter(parameter);
        return deviceFunction;
    }

    private static DeviceFunction getDeviceFunctionLed(boolean setOn) {
        DeviceFunction deviceFunction = new DeviceFunction("led", DeviceFunction.Type.ACTUATOR);
        deviceFunction.setGuiName("Turn led on/off");
        Parameter parameter = new Parameter("ToggleButton", "Boolean");
        parameter.setValueAsString(setOn ? "HIGH" : "LOW");
        deviceFunction.addParameter(parameter);
        return deviceFunction;
    }

    private static DeviceFunction getDeviceFunctionLedFlash(int timer) {
        DeviceFunction deviceFunction = new DeviceFunction("flash", DeviceFunction.Type.ACTUATOR);
        deviceFunction.setGuiName("Turn led on for x seconds");
        Parameter parameter = new Parameter("Slider", "Integer");
        parameter.setValueAsString(timer+"");
        deviceFunction.addParameter(parameter);
        deviceFunction.setContinuous(true);
        return deviceFunction;
    }

    private static DeviceFunction getDeviceFunctionRfidRead(RFIDGroup group) {
        DeviceFunction deviceFunction = new DeviceFunction("rfidread", DeviceFunction.Type.SENSOR);
        deviceFunction.setGuiName("Read RFID tag");
        Parameter parameter = new Parameter();
        parameter.setValueAsString(group.getRFIDTags());
        parameter.setLabel(group.getName());
        parameter.setUiElementName("RFID");
        parameter.setDatatypeName("String");
        deviceFunction.addParameter(parameter);
        return deviceFunction;
    }

    private static DeviceFunction getDeviceFunctionSpeakerMario() {
        DeviceFunction deviceFunction = new DeviceFunction("melody2", DeviceFunction.Type.ACTUATOR);
        deviceFunction.setGuiName("Play Mario theme");
        deviceFunction.setContinuous(true);
        return deviceFunction;
    }

    private static DeviceFunction getDeviceFunctionSpeakerJingle() {
        DeviceFunction deviceFunction = new DeviceFunction("melody4", DeviceFunction.Type.ACTUATOR);
        deviceFunction.setGuiName("Play Jingle Bells");
        deviceFunction.setContinuous(true);
        return deviceFunction;
    }

    private static DeviceFunction getDeviceFunctionTilt() {
        DeviceFunction deviceFunction = new DeviceFunction("tilt", DeviceFunction.Type.SENSOR);
        deviceFunction.setGuiName("Tilted");
        return deviceFunction;
    }

    private static DeviceFunction getDeviceFunctionButtonPressed() {
        DeviceFunction deviceFunction = new DeviceFunction("buttonPressed", DeviceFunction.Type.SENSOR);
        deviceFunction.setGuiName("Button click");
        return deviceFunction;
    }

    private static Condition getCondition(DeviceFunction deviceFunction, Device device, Board board) {
        Condition condition = new Condition(deviceFunction, board);
        condition.setDevice(device);
        return condition;
    }

    private static AutomaticSendMessage getAutomaticSendMessage(Device device, Board board) {
        AutomaticSendMessage action = new AutomaticSendMessage(board);
        action.setDevice(device);
        return action;
    }

    private static AutomaticReceiveMessage getAutomaticReceiveMessage(Device device, Board board) {
        AutomaticReceiveMessage automaticReceiveMessage = new AutomaticReceiveMessage(board);
        automaticReceiveMessage.setDevice(device);
        return automaticReceiveMessage;
    }

    private static Action getAction(DeviceFunction deviceFunction, Device device, Board board) {
        Action action = new Action(deviceFunction, board);
        action.setDevice(device);
        return action;
    }
}