/*
 * Licensed to UbiCollab.org under one or more contributor
 * license agreements.  See the NOTICE file distributed
 * with this work for additional information regarding
 * copyright ownership. UbiCollab.org licenses this file
 * to you under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package no.ntnu.android.toolkit.bluetooth;

import android.app.Activity;
import android.app.ProgressDialog;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.content.BroadcastReceiver;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.IntentFilter;
import android.view.Gravity;
import android.view.View;
import android.widget.*;
import no.ntnu.android.toolkit.activities.BoardBluetoothMapping;
import no.ntnu.android.toolkit.activities.ConfigureRules;
import no.ntnu.android.toolkit.activities.MyActivity;
import no.ntnu.android.toolkit.graphic.Constant;
import no.ntnu.android.toolkit.graphic.adapter.BoardBluetoothAdapter;
import no.ntnu.android.toolkit.model.Board;
import no.ntnu.android.toolkit.model.HardwareSetup;
import no.ntnu.android.toolkit.model.Model;
import no.ntnu.android.toolkit.reader.BoardTypeReader;
import no.ntnu.android.toolkit.tool.FormatCheck;

import java.util.*;

public class DiscoverDevices {
    private BluetoothAdapter bluetoothAdapter;
    private Activity activity;
    private int adapterResourceType;
    private BluetoothListAdapter adapter;
    private List<BluetoothDevice> pairedBluetoothDeviceList;
    private String fromActivity;
    private BroadcastReceiver broadcastReceiver;
    private ListView bluetoothList;
    private LinearLayout linearLayout;
    private HashMap<String, String> checkedDevices;
    private BoardTypeReader boardTypeReader;
    private List<HardwareSetup> hardwareSetups;
    private DiscoverDevicesProgressDialogHandler discoverDevicesProgressDialogHandler;
    private BoardBluetoothAdapter boardBluetoothAdapter;
    private ProgressDialog currentProgressDialog;
    private static int searchLimit = 2;
    private int searchCount = 0;
    private Intent intent;

    public DiscoverDevices(Activity activity, LinearLayout linearLayout, int  adapterResourceType) {
        this.activity = activity;
        Intent receivedIntent = activity.getIntent();
        fromActivity = receivedIntent.getStringExtra("fromActivity");
        this.linearLayout = linearLayout;
        this.adapterResourceType = adapterResourceType;
        bluetoothList = new ListView(activity);
        adapter = new BluetoothListAdapter(activity, adapterResourceType, android.R.id.text1);
        bluetoothList.setAdapter(adapter);
        checkedDevices = new HashMap<String, String>();
        boardTypeReader = new BoardTypeReader(activity);
        hardwareSetups = boardTypeReader.getDevices();
    }

    public void showSearchDialog() {
        if(activity instanceof ConfigureRules) {
            try {
                currentProgressDialog.dismiss();
            } catch (Exception e) {
                //Do nothing
            }
            final ConfigureRules instance = (ConfigureRules)activity;
            discoverDevicesProgressDialogHandler = new DiscoverDevicesProgressDialogHandler((ConfigureRules)activity);
            currentProgressDialog = discoverDevicesProgressDialogHandler.createSearchDialog();
            currentProgressDialog.show();
            currentProgressDialog.getButton(DialogInterface.BUTTON_NEGATIVE).setOnClickListener(new View.OnClickListener() {
                @Override
                public void onClick(View view) {
                    cancelDiscovery();
                    instance.onBackPressed();
                }
            });
        }
        initiateSearch();
    }

    public DiscoverDevices(Activity activity, BoardBluetoothAdapter boardBluetoothAdapter, LinearLayout linearLayout, Intent intent) {
        this.activity = activity;
        this.boardBluetoothAdapter = boardBluetoothAdapter;
        this.linearLayout = linearLayout;
        bluetoothList = new ListView(activity);
        bluetoothList.setAdapter(boardBluetoothAdapter);
        this.intent = intent;
    }

    public void initiateSearch() {
        enableBluetooth();
        if (bluetoothAdapter.isDiscovering()) {
            cancelOngoingSearch();
        }
        if (bluetoothAdapter.isEnabled()) {
            searchCount++;
            linearLayout.removeAllViews();
            if (adapter != null) {
                adapter.clear();
            } else {
                boardBluetoothAdapter.clear();
            }
            TextView textViewNoDevice = new TextView(activity);
            textViewNoDevice.setText("Searching for paired bluetooth boards...");
            textViewNoDevice.setTextSize(15);
            textViewNoDevice.setGravity(Gravity.CENTER);
            textViewNoDevice.setPadding(0, 20, 0, 0);
            linearLayout.addView(textViewNoDevice);
            getPairedDevices();
            populateBluetoothList();
        }

    }

    private void cancelOngoingSearch(){
        bluetoothAdapter.cancelDiscovery();
        unregisterBroadcastReceiver();
    }

    private void enableBluetooth(){
        bluetoothAdapter = BluetoothAdapter.getDefaultAdapter();
        if (!bluetoothAdapter.isEnabled()) {
            Intent enableBtIntent = new Intent(BluetoothAdapter.ACTION_REQUEST_ENABLE);
            activity.startActivityForResult(enableBtIntent, Constant.ENABLE_BT_REQUEST);
        }
    }

    private void getPairedDevices(){
        pairedBluetoothDeviceList = new ArrayList<BluetoothDevice>();
        if(adapterResourceType == android.R.layout.simple_list_item_1){
            bluetoothList.setOnItemClickListener(new AdapterView.OnItemClickListener() {
                @Override
                public void onItemClick(AdapterView<?> adapterView, View view, int i, long l) {
                    String s =  (String) bluetoothList.getItemAtPosition(i);
                    String[] split = s.split("\n");
                    boolean isValid = FormatCheck.validateMacAdress((split[1]));

                    if(fromActivity.equalsIgnoreCase("ConfigureRules")){
                        Model.getInstance().getActiveBoard().setBluetoothMacAddress(split[1]);
                        activity.setResult(Activity.RESULT_OK);
                        activity.finish();
                        unregisterBroadcastReceiver();
                    }
                    else if(fromActivity.equalsIgnoreCase("MyActivity")){
                        Intent redirect = new Intent(activity.getApplicationContext(), MyActivity.class);
                        activity.startActivity(redirect);
                        unregisterBroadcastReceiver();
                        if(isValid) {
                            Toast.makeText(activity.getApplicationContext(), "Default bluetooth device is now set", Toast.LENGTH_SHORT).show();
                        }
                    }
                }
            });
        } else if(adapterResourceType == android.R.layout.simple_list_item_multiple_choice){
            bluetoothList.setOnItemClickListener(new AdapterView.OnItemClickListener() {
                @Override
                public void onItemClick(AdapterView<?> adapterView, View view, int i, long l) {
                    if(((CheckedTextView)view).isChecked()){
                        ((CheckedTextView)view).setChecked(false);
                        checkedDevices.remove(addSelectedDevice((String) ((CheckedTextView) view).getText()));
                    } else {
                        ((CheckedTextView)view).setChecked(true);
                        addSelectedDevice((String) ((CheckedTextView) view).getText());
                    }
                }
            });
        }
        Set<BluetoothDevice> pairedDevices = bluetoothAdapter.getBondedDevices();
        pairedBluetoothDeviceList.addAll(pairedDevices);
    }
    private void populateBluetoothList(){
        bluetoothAdapter.startDiscovery();

        broadcastReceiver = new BroadcastReceiver() {
            @Override
            public void onReceive(android.content.Context context, Intent intent) {
                String action = intent.getAction();
                if (BluetoothDevice.ACTION_FOUND.equals(action)) {
                    linearLayout.removeAllViews();
                    if (adapter == null) {
                        ((BoardBluetoothMapping)activity).addMenuHeader(linearLayout);
                    }
                    linearLayout.addView(bluetoothList);
                    BluetoothDevice device = intent.getParcelableExtra(BluetoothDevice.EXTRA_DEVICE);
                    if(pairedBluetoothDeviceList.contains(device)) {
                        if (adapter != null) {
                            adapter.add(device.getName() + "\n" + device.getAddress());
                        } else{
                            boardBluetoothAdapter.add(device);
                        }
                        if (activity instanceof ConfigureRules) {
                            for (HardwareSetup hardwareSetup : hardwareSetups) {
                                boolean boardUnique = true;
                                if (hardwareSetup.getName().equalsIgnoreCase(device.getName())) {
                                    for (Board board : Model.getInstance().getBoards()) {
                                        if (board.getName().equalsIgnoreCase(hardwareSetup.getName())) {
                                            boardUnique = false;
                                        }
                                    }
                                    if(boardUnique){
                                        Board board = Model.getInstance().createAddBoard(hardwareSetup);
                                        board.setBluetoothMacAddress(device.getAddress());
                                        discoverDevicesProgressDialogHandler.deviceDiscoveredChange(DiscoverDevicesProgressDialogHandler.EventType.DEVICE_ADDED);
                                        searchingProgressDialogChangeEvent(DiscoverDevicesProgressDialogHandler.EventType.DEVICE_ADDED);
                                        BoardTypeReader.addDeviceFunctionsToBoards(activity);
                                        if (board == Model.getInstance().getActiveBoard()) {
                                            ((ConfigureRules) activity).initiateBoardItems();
                                        }
                                    }
                                }
                            }
                        }
                    }
                    if (adapter != null) {
                        adapter.notifyDataSetChanged();
                    } else{
                        boardBluetoothAdapter.notifyDataSetChanged();
                    }
                    if (adapter == null) {
                        if (boardBluetoothAdapter.getCount() > 0) {
                            createSaveButtonInBoardMapping();
                        }

                    }
                } else if (BluetoothAdapter.ACTION_DISCOVERY_FINISHED.equals(action)) {
                    if(searchCount >= searchLimit) {
                        searchCount = 0;
                        currentProgressDialog.setMessage("No devices found");
                        currentProgressDialog.setIndeterminateDrawable(null);
                        currentProgressDialog.getButton(DialogInterface.BUTTON_NEGATIVE).setText("Ok");
                    } else if (Model.getInstance().getActiveBoard() == null) {
                        initiateSearch();
                    }
                }

            }
        };
        unregisterBroadcastReceiver();
        IntentFilter found_filter = new IntentFilter(BluetoothDevice.ACTION_FOUND);
        IntentFilter finished_filter = new IntentFilter(BluetoothAdapter.ACTION_DISCOVERY_FINISHED);
        activity.registerReceiver(broadcastReceiver, found_filter);
        activity.registerReceiver(broadcastReceiver, finished_filter);
    }

    public BluetoothAdapter getBluetoothAdapter(){
        return bluetoothAdapter;
    }
    public HashMap<String, String> getCheckedDevices(){
        return checkedDevices;
    }

    private String addSelectedDevice(String string){
        String [] array = string.split("\n");
        try{
            checkedDevices.put(array[0], array[1]);
        } catch (NullPointerException n){
            n.printStackTrace();
        }
        return null;
    }

    public void cancelDiscovery(){
        if(bluetoothAdapter.isDiscovering()){
            cancelOngoingSearch();
        }
    }

    public void unregisterBroadcastReceiver(){
        try{
            if(broadcastReceiver != null){
                activity.unregisterReceiver(broadcastReceiver);
            }
        } catch (Exception e){
            //Do nothing
        }
    }

    public void createBoardsAndAddDeviceFunctions(HashMap<String, String> bluetoothDeviceNames){
        for (HardwareSetup hardwareSetup : hardwareSetups) {
            for(String bluetoothDeviceName : bluetoothDeviceNames.keySet()){
                if(hardwareSetup.getName().equalsIgnoreCase(bluetoothDeviceName)){
                    Model.getInstance().createAddBoard(hardwareSetup);
                    Model.getInstance().getActiveBoard().setBluetoothMacAddress(bluetoothDeviceNames.get(bluetoothDeviceName));
                    BoardTypeReader.addDeviceFunctionsToBoards(activity);
                }
            }
        }
        if(Model.getInstance().getBoards().size() == 0){
            Toast.makeText(activity, "No corresponding board types found. Are you sure it is an Arduino?", Toast.LENGTH_LONG).show();
        } else{
            Intent redirectToConfigureRules = new Intent(activity, ConfigureRules.class);
            activity.startActivity(redirectToConfigureRules);
        }
    }

    private void searchingProgressDialogChangeEvent(final DiscoverDevicesProgressDialogHandler.EventType type) {
        activity.runOnUiThread(new Runnable() {
            @Override
            public void run() {
                discoverDevicesProgressDialogHandler.deviceDiscoveredChange(type);
            }
        });
    }
    private void createSaveButtonInBoardMapping() {

        Button saveButton = new Button(activity);
        saveButton.setText("Save");
        linearLayout.addView(saveButton);

        saveButton.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View view) {
                List<Board> selectedBoards = new ArrayList<Board>();

                //Prepare yourself: This is some nasty piece of code
                LinearLayout boardMappingLayout = (LinearLayout) view.getParent();
                for (int i = 0; i < boardMappingLayout.getChildCount(); i++) {
                    if (boardMappingLayout.getChildAt(i) instanceof ListView) {
                        ListView list = (ListView) boardMappingLayout.getChildAt(i);
                            for (int e = 0; e < list.getChildCount(); e++) {
                                BoardBluetoothAdapter.BoardBluetoothListItem boardBluetoothListItem = (BoardBluetoothAdapter.BoardBluetoothListItem) list.getItemAtPosition(e);
                                TableLayout tableLayout = (TableLayout) list.getChildAt(e);
                                for (int c = 0; c < tableLayout.getChildCount(); c++) {
                                    TableRow row = (TableRow) tableLayout.getChildAt(c);
                                    for (int d = 0; d < row.getChildCount(); d++) {
                                        if (row.getChildAt(d) instanceof Spinner) {
                                            Spinner spinner = (Spinner) row.getChildAt(d);
                                            Board board = (Board) spinner.getSelectedItem();
                                            if (!board.getName().equalsIgnoreCase("Select board")) {
                                                board.setBluetoothMacAddress(boardBluetoothListItem.getBluetoothDeviceMac());
                                                selectedBoards.add(board);
                                            }
                                        }
                                    }
                                }
                            }
                    }
                }

                HashMap<Board, Integer> spinnerSelectedItems = new HashMap<Board, Integer>();
                for (Board board : selectedBoards) {
                    System.out.println(board);
                    if (!spinnerSelectedItems.containsKey(board)) {
                        spinnerSelectedItems.put(board, 1);
                    } else{
                        spinnerSelectedItems.put(board, spinnerSelectedItems.get(board) + 1);
                    }
                }

                boolean boardSelectedMoreThanOnce = true;
                Iterator it = spinnerSelectedItems.entrySet().iterator();
                while (it.hasNext()) {
                    Map.Entry pairs = (Map.Entry)it.next();
                    if ((Integer)pairs.getValue() > 1) {
                        boardSelectedMoreThanOnce = false;
                    }
                }

                if (intent.getStringExtra("fromActivity") != null){
                    System.out.println("BoardSelectedMoreTHanOne: " + boardSelectedMoreThanOnce);
                    if((intent.getStringExtra("fromActivity").equalsIgnoreCase("menu") || intent.getStringExtra("fromActivity").equalsIgnoreCase("upload")) && boardSelectedMoreThanOnce){
                        activity.setResult(Activity.RESULT_OK);
                        ((BoardBluetoothMapping)activity).done();
                    }
                     else{
                        Toast.makeText(activity, "One board cannot have several devices", Toast.LENGTH_SHORT).show();
                    }
                }else {
                    Toast.makeText(activity, "One board cannot have several devices", Toast.LENGTH_SHORT).show();
                }
            }

        });

    }
}

