/*
 * Licensed to UbiCollab.org under one or more contributor
 * license agreements.  See the NOTICE file distributed
 * with this work for additional information regarding
 * copyright ownership. UbiCollab.org licenses this file
 * to you under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package no.ntnu.android.toolkit.generator;

import android.content.res.AssetManager;
import no.ntnu.android.toolkit.model.*;
import org.apache.commons.lang3.StringUtils;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class Generator {

    private Map<String, ArrayList<String>> extraParameters = new HashMap<String, ArrayList<String>>();
    private final ArrayList<Rule> rules;
    private final AssetManager assetManager;
    private ArrayList<Device> devicesHavingContinuous = new ArrayList<Device>();
    private Board board;

    public Generator(final ArrayList<Rule> rules, final AssetManager assetManager) {
        this.board = rules.get(0).getBoard();
        this.rules = rules;
        this.assetManager = assetManager;
    }

    private String classes() {
        boolean continuousFound = false;
        String classesString = "";
        for (Rule rule : rules) {
            for (Action action : rule.getActions()) {
                if (action.getDeviceFunction().isContinuous()) {
                    classesString = "class Continuous { public: boolean b; Continuous(boolean a){ b = a;} };\n";
                    continuousFound = true;
                    break;
                }
            }
            if (continuousFound) {
                break;
            }
        }
        for (Rule rule : rules) {
            for (Configuration configuration : rule.getConfigurations()) {
                String classText = getCodeFromFile(configuration, rule, "classes");
                if (!classText.equals("")) {
                    List<String> classFunctions = splitClasses(classText);
                    for(String function : classFunctions){
                        if (!classesString.contains(function)) {
                            classesString += function + "\n";
                        }
                    }
                }
            }
        }
        return classesString;
    }

    private List<String> splitClasses(String classText){
        List<String> classFunctions = new ArrayList<String>();
        String text = classText;
        int startValue = 0;
        int numberOfMatches = StringUtils.countMatches(classText, "class");
        int added = 0;
        if(numberOfMatches> 0){
            while(added<= numberOfMatches){
                int firstInstance = StringUtils.indexOf(text, "class", startValue);
                classFunctions.add(StringUtils.substring(classText, startValue, firstInstance).trim());
                added++;
                text = StringUtils.substring(classText, firstInstance);
                startValue = firstInstance;
            }
            return classFunctions;
        }
        return null;
    }

    private String imports() {
        String importString = "";
        if (board.getWifiDevice() != null) {
            importString += "#include <WiFlyHQ.h>\n";
        }
        return importString + "#include <ComputerSerial.h>\n#include <SoftwareSerial.h>";
    }

    private String pinVariables() {
        ArrayList<String> createdVariables = new ArrayList<String>();
        String variableString = "";
        for (Rule rule : rules) {
            for (Configuration configuration : rule.getConfigurations()) {
                ArrayList<String> configurationVariables = configurationPinVariables(configuration);
                int i = 0;
                for (String variable : configurationVariables) {
                    if (!createdVariables.contains(variable)) {
                        variableString += "int " + variable + " = "
                                + configuration.getDevice().getPins().get(i).getNumber() + ";";
                        createdVariables.add(variable);
                    }
                    i++;
                }
            }
        }

        return variableString;
    }

    private String continuousVariables() {
        String variables = "\n";

        for (Rule rule : rules) {
            for (Action action : rule.getActions()) {
                if (action.getDeviceFunction().isContinuous()) {
                    String newVariable = continuousVariable(action);
                    if (!variables.contains(newVariable)) {
                        variables += "Continuous " + newVariable + " = Continuous(false);";
                        devicesHavingContinuous.add(action.getDevice());
                    }
                }

            }
        }
        return variables;
    }

    private String variables() {
        String variablesString = pinVariables() + getVariableDeclarationsFromFile() + continuousVariables();
        variablesString += "static ComputerSerial comp;";
        if (board.getWifiDevice() != null) {
            variablesString += "\nboolean wifiEnable = false;\n";
            variablesString += "SoftwareSerial wifiSerial(" + board.getWifiDevice().getPins().get(0).getNumber() +
                ", " + board.getWifiDevice().getPins().get(1).getNumber() + ");\n";
            variablesString += "WiFly wifly;\n";
        }
        return variablesString;
    }

    private String getVariableDeclarationsFromFile() {
        String variableDec = "";
        ArrayList<String> alreadyCreated = new ArrayList<String>();
        for (Rule rule : rules) {
            for (Configuration configuration : rule.getConfigurations()) {
                String vardec = getCodeFromFile(configuration, rule, "variables");
                for (String dec : vardec.split(";")) {
                    if (!alreadyCreated.contains(dec) && !dec.equals(vardec.split(";")[vardec.split(";").length - 1])) {
                        if (!dec.contains("SoftwareSerial") || !variableDecContainsSoftwareSerialVariable(variableDec, dec)) {
                            variableDec += dec + ";";
                            alreadyCreated.add(dec);
                        }
                    }
                }
            }
        }
        return variableDec;
    }

    private boolean variableDecContainsSoftwareSerialVariable(String variableDec, String variableLine) {
        String variable = variableLine.split("\\(")[0].split(" ")[1];
        return variableDec.contains(variable);
    }

    private ArrayList<String> configurationPinVariables(Configuration configuration) {
        ArrayList<String> variables = new ArrayList<String>();
        for (Pin pin : configuration.getDevice().getPins()) {
            variables.add(configuration.getDeviceFunction().getName() + configuration.getDevice().getId() +"_" + pin.getNumber());
        }
        return variables;
    }

    private String setup() {
        if (rules == null || rules.size() == 0) {
            return "void setup(){\ncomp.begin(115200);}";
        }
        String setupString = "void setup(){\ncomp.begin(115200);\n";
        if (board.getWifiDevice() != null) {
            setupString += "wifiSerial.begin(9600);\n";
            setupString += "wifly.begin(&wifiSerial);\n";
        }
        if (board.rulesRequireWifi()) {
            setupString += wifiSetup();
        }
        ArrayList<String> createdConfigurationSetups = new ArrayList<String>();
        for (Rule rule : rules) {
            for (Configuration configuration : rule.getConfigurations()) {
                String configurationSetup = configrationSetup(configuration);
                if (!createdConfigurationSetups.contains(configurationSetup)) {
                    setupString += configurationSetup;
                    createdConfigurationSetups.add(configurationSetup);
                }
            }
        }
        setupString += "}";
        return setupString;
    }

    private String wifiSetup() {
        String setupString = "\n";
        setupString += "if (!wifly.isAssociated()) {\n";
        setupString += "wifly.setSSID(\"" + board.getWifiSettings().getSsid() + "\");\n";
        setupString += "wifly.setPassphrase(\"" + board.getWifiSettings().getPassword() + "\");\n";
        setupString += "wifly.enableDHCP();\n";
        setupString += "if (wifly.join()) {\n";
        setupString += "wifiEnable = true;\n}\n";
        setupString += "} else {\n";
        setupString += "wifiEnable = true;\n}\n";
        setupString += "wifly.setIpProtocol(WIFLY_PROTOCOL_UDP);\n";
        return setupString;
    }

    private String configrationSetup(Configuration configuration) {
        String setup = getCodeFromFile(configuration, null, "setup");
        ArrayList<String> pinVariables = configurationPinVariables(configuration);
        int i = 0;
        for (String variable : pinVariables) {
            String pinMode = configuration.getDevice().getPins().get(i).getMode().toString();
            if (!pinMode.equalsIgnoreCase("NONE")) {
                setup += "pinMode(" + variable + ", " + configuration.getDevice().getPins().get(i).getMode().toString() + ");";
            }
            i++;
        }
        return setup;
    }

    private String loop() {
        String loopString = "void loop(){\n";
        loopString += functionLoopPart();
        loopString += continuousLoopPart() + "\n";
        loopString += ifLoopPart();
        return loopString + "\n}";
    }

    private String functionLoopPart() {
        String loopString = "";
        for (Rule rule : rules) {
            for (Configuration configuration : rule.getConfigurations()) {
                String code = getCodeFromFile(configuration, rule, "loop");
                if (!loopString.contains(code)) {
                    loopString += code;
                }
            }
        }
        return loopString;
    }

    private String ifLoopPart() {
        String loopString = "";
        for (Rule rule : rules) {
            loopString += ifStatement(rule) + "{\n";
            for (Action action : rule.getActions()) {
                if (devicesHavingContinuous.contains(action.getDevice())) {
                    loopString += clearFunctionName(action.getDevice()) + "();\n";
                }
            }
            for (Action action : rule.getActions()) {
                if (!action.getDeviceFunction().isContinuous()) {
                    loopString +=  functionCall(action, rule) + ";";
                }
                else {
                    loopString += continuousVariable(action) + ".b = true; ";
                }
                loopString += "\n";
            }
            loopString += "}";
        }
        return loopString;
    }

    private String clearFunctionName(Device device) {
        return "clear" + device.getType() + device.getId();
    }

    private String continuousLoopPart() {
        String calls = "";
        for (Rule rule : rules) {
            for (Action action : rule.getActions()) {
                if (action.getDeviceFunction().isContinuous()) {
                    calls += "if (" + continuousVariable(action) + ".b == true) {\n" +
                            functionCall(action, rule) + ";\n}";
                }
            }
        }
        return calls + "\n";
    }

    private String continuousVariable(Action action) {
        return action.getDeviceFunction().getName() + action.getId() + "Continuous";
    }



    private String ifStatement(Rule rule) {
        String ifString = "if (";
        int i = 0;
        for (Condition condition : rule.getConditions()) {
            ifString += functionCall(condition, rule);
            if (rule.getOperators().size() > i) {
                ifString += " " + rule.getOperators().get(i) + " ";
                i += 1;
            }
        }
        ifString += ")";
        return ifString;
    }

    private String functionCall(Configuration configuration, Rule rule) {
        String functionCall = configuration.getDeviceFunction().getName() + "(";
        ArrayList<String> pinVariables = configurationPinVariables(configuration);
        int i = 0;
        for (String variable : pinVariables) {
            String pinMode = configuration.getDevice().getPins().get(i).getMode().toString();
            if (!pinMode.equalsIgnoreCase("NONE")) {
                functionCall += variable + ", ";
            }
            i++;
        }
        for (Parameter parameter : configuration.getDeviceFunction().getParameters()) {
            if (parameter.getDatatypeName().equalsIgnoreCase("string")) {
                functionCall += "\"" + parameter.getValueAsString() + "\", ";
            } else {
                functionCall += parameter.getValueAsString() + ", ";
            }
        }
        if (extraParameters.get(rule.getId() + "-" + configuration.getId()) != null) {
            for(String variable : extraParameters.get(rule.getId() + "-" + configuration.getId())){
                if(extraParameters.get(rule.getId() + "-" + configuration.getId()).
                        get(extraParameters.get(rule.getId() + "-" + configuration.getId()).size() - 1).equals(variable)){
                    functionCall += variable;

                } else{
                    functionCall += variable + ", ";
                }
            }
            extraParameters.remove(rule.getId() + "-" + configuration.getId());
        }
        if (functionCall.substring(functionCall.length() - 2).equals(", ")) {
            functionCall = functionCall.substring(0, functionCall.length() - 2);
        }
        if(configuration instanceof Action){
            if(configuration.getDeviceFunction().isContinuous()){
               functionCall += ", " + continuousVariable((Action) configuration);
            }
        }
        functionCall += ")";
        return functionCall;
    }

    private String functions() {
        ArrayList<String> createdFunctions = new ArrayList<String>();
        String functionsString = "";
        for (Rule rule : rules) {
            for (Configuration configuration : rule.getConfigurations()) {
                String function = getCodeFromFile(configuration, rule, "function");
                if (!createdFunctions.contains(function)) {
                    functionsString += function;
                    createdFunctions.add(function);
                }
            }
        }
        functionsString += clearFunctions() + serialEventFunction();
        return functionsString;
    }

    private String clearFunctions() {
        String clearFunctions = "";
        ArrayList<Device> alreadyCreated = new ArrayList<Device>();
        for (Rule rule : rules) {
            for (Action action : rule.getActions()) {
                if (action.getDeviceFunction().isContinuous() && !alreadyCreated.contains(action.getDevice())) {
                    clearFunctions += clearFunction(action.getDevice()) + "\n";
                    alreadyCreated.add(action.getDevice());
                }
            }
        }
        return clearFunctions;
    }

    private String clearFunction(Device device) {
        String function = "void " + clearFunctionName(device) + "() {";
        for (Rule rule : rules) {
            for (Action action : rule.getActions()) {
                if (action.getDevice() == device && action.getDeviceFunction().isContinuous() &&
                        !function.contains(continuousVariable(action))) {
                    function += "\n" + continuousVariable(action) + ".b = false;";
                }
            }
        }
        return function + "\n}";
    }

    private String functionDeclaration() {
        String decString = "void setup();void loop();void serialEvent();";
        for (Rule rule : rules) {
            for (Configuration configuration : rule.getConfigurations()) {
                String function = configuration.getDeviceFunction().getName();
                if (!decString.contains(function)) {
                    decString += getFunctionDeclarationFromFile(configuration);
                }
            }
            for (Action action : rule.getActions()) {
                if (action.getDeviceFunction().isContinuous()) {
                    if (!decString.contains(clearFunctionName(action.getDevice()))) {
                        decString += "\nvoid " + clearFunctionName(action.getDevice()) + "();";
                    }
                }
            }
        }
        return decString;
    }

    private String getFunctionDeclarationFromFile(Configuration configuration) {
        return getCodeFromFile(configuration, null, "function").split("\\)")[0] + ");";
    }

    private String getCodeFromFile(Configuration configuration, Rule rule, String suffix) {
        if (assetManager == null) return "";
        String filePath = "configurations";
        if (configuration instanceof Condition) filePath += "/conditions/";
        else filePath += "/actions/";
        filePath += configuration.getDeviceFunction().getName() + "_" + suffix + ".txt";

        try {
            String code = readAsset(filePath);
            while (code.contains("#c[")) {
                String between = StringUtils.substringBetween(code, "#c[", "]c#");
                String variable = between + configuration.getId();
                if(suffix.equalsIgnoreCase("Variables") && !variable.contains("^")){
                    if(extraParameters.get(rule.getId() + "-" + configuration.getId()) == null){
                        extraParameters.put(rule.getId() + "-" + configuration.getId(), new ArrayList<String>());
                    }
                    extraParameters.get(rule.getId() + "-" + configuration.getId()).add(variable);
                }
                variable = StringUtils.remove(variable, '^');
                code = StringUtils.replace(code, "#c[" + between + "]c#", variable);
            }
            while (code.contains("#d[")) {
                String between = StringUtils.substringBetween(code, "#d[", "]d#");
                String variable;
                if (between.contains(":")) {
                    int pinIndex = Integer.parseInt(between.split(":")[1]);
                    variable = between.split(":")[0] + configuration.getDevice().getId() + "_" +
                            configuration.getDevice().getPins().get(pinIndex).getNumber();
                } else {
                    variable = between + configuration.getDevice().getId();
                }
                if(suffix.equalsIgnoreCase("Variables") && !variable.contains("^")){
                    if(extraParameters.get(rule.getId() + "-" + configuration.getId()) == null){
                        extraParameters.put(rule.getId() + "-" + configuration.getId(), new ArrayList<String>());
                    }
                    extraParameters.get(rule.getId() + "-" + configuration.getId()).add(variable);
                }
                variable = StringUtils.remove(variable, '^');
                code = StringUtils.replace(code, "#d[" + between + "]d#", variable);
            }
            return code;
        } catch (IOException e) {
            return "";
        }
    }

    private String readAsset(String filePath) throws IOException {
        String functionString = "";
        BufferedReader br = new BufferedReader(new InputStreamReader(assetManager.open(filePath)));
        String line = "";
        do {
            functionString += line + "\n";
            line = br.readLine();
        } while (line != null);

        return functionString;
    }

    private String serialEventFunction() {
        if (board.getWifiDevice() != null) {
            return "void serialEvent() {\n" +
                    "  if (Serial.peek() == 0xFD) {\n" +
                    "      delay(1000);\n" +
                    "      Serial.read();\n" +
                    "      boolean ssidDone = false;\n" +
                    "      String ssid = \"\";\n" +
                    "      String pass = \"\";\n" +
                    "      while (Serial.available() > 0) {\n" +
                    "        if (ssidDone) {\n" +
                    "           pass = pass + (char)Serial.read();\n" +
                    "        } else if (((char)Serial.peek()) == ':') {\n" +
                    "           Serial.read();\n" +
                    "           ssidDone = true;\n" +
                    "        } else {\n" +
                    "          ssid = ssid + (char)Serial.read(); \n" +
                    "        }\n" +
                    "      }\n" +
                    "      wifiSerial.listen();\n" +
                    "        if (wifly.isAssociated()) {\n" +
                    "            wifly.leave();\n}\n" +
                    "        delay(1000);\n" +
                    "        char ssidc[64];\n" +
                    "        ssid.toCharArray(ssidc, sizeof(ssidc));\n" +
                    "        wifly.setSSID(ssidc);\n" +
                    "        char passc[64];\n" +
                    "        pass.toCharArray(passc, sizeof(passc));\n" +
                    "        wifly.setPassphrase(passc);\n" +
                    "        wifly.enableDHCP();\n" +
                    "        if (wifly.join()) {\n" +
                    "          char buff[64];\n" +
                    "          wifly.getIP(buff, sizeof(buff));\n" +
                    "          Serial.print(buff);\n" +
                    "          Serial.write(0xFE);\n" +
                    "        } else {\n" +
                    "          Serial.print(\"null\");\n" +
                    "          Serial.write(0xFE);\n" +
                    "        }\n" +
                    "  } else {\n" +
                    "    comp.serialEvent();\n" +
                    "  }\n" +
                    "}";
        } else {
            return "void serialEvent() {" +
                    "if (Serial.peek() == 0xFD) {\n" +
                    "Serial.read();\n" +
                    "Serial.print(\"not_supported\");\n " +
                    "Serial.write(0xFE);\n}\n" +
                    "else {comp.serialEvent();" +
                    "}\n}";
        }
    }

    public String generateCppCode(){
        String code = "#line 1 \"Sketch.ino\"";
        code += "\n" + imports();
        code += "\n#include \"Arduino.h\"";
        code += "\n" + functionDeclaration();
        code += "\n#line 2";
        code += "\n" + classes();
        code += "\n" + variables();
        code += "\n" + setup();
        code += "\n" + loop();
        code += "\n" + functions();
        return code;
    }
}
