/*
 * Licensed to UbiCollab.org under one or more contributor
 * license agreements.  See the NOTICE file distributed
 * with this work for additional information regarding
 * copyright ownership. UbiCollab.org licenses this file
 * to you under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package no.ntnu.android.toolkit.graphic;

import android.app.Activity;
import android.app.AlertDialog;
import android.content.DialogInterface;
import android.view.View;
import android.view.ViewGroup;
import android.widget.*;
import no.ntnu.android.toolkit.activities.R;
import no.ntnu.android.toolkit.model.*;
import no.ntnu.android.toolkit.tool.Search;

import java.util.ArrayList;
import java.util.List;

public class ElementConnector {

   public static int getPictureId(Activity activity, Device device) {
       String name = device.getType();
       String nameWithText = null;
       if(device.getText() != null && !device.getText().equals("")) {
           if (device.getText().contains(":")) {
               nameWithText = device.getType() + "_" + device.getText().split(":")[0];
           } else {
               nameWithText = device.getType() + "_" + device.getText();
           }
       }
       int drawableID;
       if (nameWithText != null) {
           drawableID = activity.getResources().getIdentifier(nameWithText.toLowerCase(), "drawable", activity.getPackageName());
           if (drawableID > 0 ) {
               return drawableID;
           }
       }
       drawableID = activity.getResources().getIdentifier(name.toLowerCase(), "drawable", activity.getPackageName());
       if(drawableID > 0 ) {
           return drawableID;
       }
       throw new UnsupportedOperationException("Type:" + device.getType() + " has no connected GUI element");
   }

    public static LinearLayout getFullLayoutForConfiguration(Configuration configuration, Activity activity) {
        LinearLayout layout = new LinearLayout(activity);
        layout.setOrientation(LinearLayout.VERTICAL);
        layout.addView(getSelectionLayout(configuration, activity));
        return layout;
    }

    public static void addSaveBuilder(AlertDialog.Builder builder) {
        builder.setPositiveButton(R.string.save, new DialogInterface.OnClickListener() {
            @Override
            public void onClick(DialogInterface dialogInterface, int i) {
                dialogInterface.dismiss();
            }
        });
    }

    public static LinearLayout getSelectionLayout(final Configuration configuration, final Activity activity) {
        final ArrayAdapter<String> adapter = new ArrayAdapter<String>(activity, android.R.layout.simple_spinner_item);
        adapter.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);
        final LinearLayout linearLayout = new LinearLayout(activity);
        linearLayout.setOrientation(LinearLayout.VERTICAL);

        LinearLayout linearLayoutForSpinner = new LinearLayout(activity);
        linearLayoutForSpinner.setOrientation(LinearLayout.VERTICAL);

        final LinearLayout linearLayoutParameters = new LinearLayout(activity);
        linearLayoutParameters.setOrientation(LinearLayout.VERTICAL);

        linearLayout.addView(linearLayoutForSpinner);
        linearLayout.addView(linearLayoutParameters);

        final Spinner spinner = new Spinner(activity);
        linearLayoutForSpinner.addView(spinner);
        spinner.setAdapter(adapter);
        spinner.setLayoutParams(new TableLayout.LayoutParams(LinearLayout.LayoutParams.WRAP_CONTENT, LinearLayout.LayoutParams.WRAP_CONTENT, 1f));

        //Populate spinner
        for(DeviceFunction deviceFunction : configuration.getDevice().getDeviceFunctions()) {
            if ((configuration instanceof Condition && deviceFunction.getType() == DeviceFunction.Type.SENSOR)
                    || ((configuration instanceof Action && deviceFunction.getType() == DeviceFunction.Type.ACTUATOR))) {
                adapter.add(deviceFunction.getGuiName());
                adapter.notifyDataSetChanged();
            }
        }

        if(configuration.getDeviceFunction() != null) {
            spinner.setSelection(adapter.getPosition(configuration.getDeviceFunction().getGuiName()));
        } else {
            spinner.setSelection(0);
        }

        spinner.setOnItemSelectedListener(new AdapterView.OnItemSelectedListener() {
            @Override
            public void onItemSelected(AdapterView<?> adapterView, View view, int i, long l) {
                linearLayoutParameters.removeAllViews();
                //Keep old values if exists and same type
                if(spinner.getSelectedItem() != null ) {
                    if(configuration.getDeviceFunction() != null
                            && (spinner.getSelectedItem().toString().equalsIgnoreCase(configuration.getDeviceFunction().getGuiName()))) {

                        if(configuration.getDeviceFunction().getParameters() != null) {
                            for (int k = 0; k < configuration.getDeviceFunction().getParameters().size(); k++) {
                                String value = configuration.getDeviceFunction().getParameters().get(k).getValueAsString();
                                if(value != null && !value.equals("")) {
                                    configuration.getDeviceFunction().getParameters().get(k).setValueAsString(value);
                                }
                            }
                        }
                    } else {
                        configuration.setDeviceFunction(configuration.getDevice().createNewDeviceFunctionByName(
                                spinner.getSelectedItem().toString()));
                    }
                }
                List<View> views = getParameterElements(configuration.getDeviceFunction(),activity, configuration.getId());
                for (View parameterView : views) {
                    if(parameterView.getParent() != null) {
                        ((ViewGroup)parameterView.getParent()).removeView(parameterView);
                    }
                    linearLayoutParameters.addView(parameterView);
                }
            }
            @Override
            public void onNothingSelected(AdapterView<?> adapterView) {
            }
        });

        return linearLayout;
    }

    public static List<View> getParameterElements(DeviceFunction deviceFunction, Activity activity, int id) {
        List<View> viewElements = new ArrayList<View>();
        for (int i = 0; i < deviceFunction.getParameters().size(); i++) {
            viewElements.add(getViewByParameter(deviceFunction.getParameters().get(i), activity, id, i));
        }
        return viewElements;
    }

    private static View getViewByParameter(Parameter parameter, Activity activity, int id, int parameterNr) {
        View view;
        String initialValue;
        if((parameter.getUiElementName().equalsIgnoreCase("Slider")
            || (parameter.getUiElementName().equalsIgnoreCase("seekBar")))) {
            view = ElementCreator.createSeekBar(activity, parameter.getValueAsString(), parameter.getLabel());
            initialValue = ((SeekBar)((LinearLayout)view).getChildAt(1)).getProgress() + "";

        } else if((parameter.getUiElementName().equalsIgnoreCase("Text")
            || (parameter.getUiElementName().equalsIgnoreCase("Textfield")))) {
            view = ElementCreator.createEditTextField(activity, parameter.getValueAsString(), parameter.getLabel());
            initialValue = ((EditText)view).getText().toString();

        } else if(parameter.getUiElementName().equalsIgnoreCase("ToggleButton")) {
            view = ElementCreator.createToggleButton(activity, parameter.getValueAsString());
            initialValue = ElementCreator.getHIGHLOWFromButton(((ToggleButton) view));

        } else if(parameter.getUiElementName().equalsIgnoreCase("RFID")) {
            view = ElementCreator.createRFIDSpinner(activity, parameter.getLabel());
            initialValue = ElementCreator.getSelectedItemFromSpinner((Spinner)view);
        }
        else {
            throw new UnsupportedOperationException("Parameter: " + parameter.getUiElementName() + " has no supported GUI element");
        }

        view.setId(id);
        view.setTag(parameterNr);
        saveValueAsParameter(id, initialValue, parameterNr, null);
        return view;
    }

    public static void saveValueAsParameter(int id, String value, Object parameterNr, String label) {
        try {
            int nr = (Integer)parameterNr;
            Configuration configuration = ((Configuration) Search.getObjectById(id));
            configuration.getDeviceFunction().getParameters().get(nr).setValueAsString(value);
            if (label != null && !label.equalsIgnoreCase("")) {
                configuration.getDeviceFunction().getParameters().get(nr).setLabel(label);
            }
        } catch (Exception e) {
            System.err.println("Save unsuccessful: " + e.getMessage());
        }
    }
}