/*
 * Licensed to UbiCollab.org under one or more contributor
 * license agreements.  See the NOTICE file distributed
 * with this work for additional information regarding
 * copyright ownership. UbiCollab.org licenses this file
 * to you under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package no.ntnu.android.toolkit.graphic;


import android.app.Activity;
import android.view.KeyEvent;
import android.view.View;
import android.view.WindowManager;
import android.view.inputmethod.EditorInfo;
import android.widget.*;
import no.ntnu.android.toolkit.activities.ConfigureRules;
import no.ntnu.android.toolkit.model.Model;
import no.ntnu.android.toolkit.model.RFIDGroup;

import java.text.DecimalFormat;
import java.util.List;

public class ElementCreator {

    public static LinearLayout createSeekBar(Activity activity, String startValue, final String label){
        final LinearLayout layout = new LinearLayout(activity);
        layout.setOrientation(LinearLayout.VERTICAL);
        final TextView textView = new TextView(activity);
        SeekBar seekBar = new SeekBar(activity);
        seekBar.setMax(5000);
        seekBar.setMinimumWidth(100);
        seekBar.setLayoutParams(new TableLayout.LayoutParams(LinearLayout.LayoutParams.MATCH_PARENT, LinearLayout.LayoutParams.MATCH_PARENT, 1f));
        try {
            seekBar.setProgress(Integer.parseInt(startValue));
        } catch (Exception e) {
            seekBar.setProgress(seekBar.getMax()/2);
        }
        final DecimalFormat oneZeroFormat = new DecimalFormat("0.0");
        textView.setText(label + " " + oneZeroFormat.format(((double)seekBar.getProgress())/1000)+ " seconds");
        seekBar.setOnSeekBarChangeListener(new SeekBar.OnSeekBarChangeListener() {
            @Override
            public void onProgressChanged(SeekBar seekBar, int i, boolean b) {
                textView.setText(label + " " + oneZeroFormat.format(((double)seekBar.getProgress())/1000)+ " seconds");
            }

            @Override
            public void onStartTrackingTouch(SeekBar seekBar) {
            }

            @Override
            public void onStopTrackingTouch(SeekBar seekBar) {
                ElementConnector.saveValueAsParameter(layout.getId(), seekBar.getProgress() + "", layout.getTag(), null);
            }
        });
        layout.addView(textView);
        layout.addView(seekBar);
        return layout;
    }

    public static ToggleButton createToggleButton(Activity activity, String startValue) {
        final ToggleButton toggleButton = new ToggleButton(activity);
        toggleButton.setTextOff("ON");
        toggleButton.setTextOff("OFF");
        toggleButton.setChecked(getTrueFalseValueFromHIGHLOW(startValue));
        toggleButton.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View v) {
                ElementConnector.saveValueAsParameter(toggleButton.getId(), getHIGHLOWFromButton(toggleButton), toggleButton.getTag(), null);
            }
        });
        return toggleButton;
    }

    public static String getHIGHLOWFromButton(ToggleButton toggleButton) {
        if(toggleButton.isChecked()) {
            return "HIGH";
        } else {
            return "LOW";
        }
    }

    public static boolean getTrueFalseValueFromHIGHLOW(String value) {
        if(value != null) {
            return value.equalsIgnoreCase("HIGH");
        }
        return false;
    }

    public static EditText createEditTextField(final Activity activity, String startValue, String hint) {
        final EditText editText = new EditText(activity);
        if (startValue != null && !startValue.equals("")) {
            editText.setText(startValue);
        } else if (hint != null) {
            editText.setHint(hint);
        }

        editText.setOnFocusChangeListener(new View.OnFocusChangeListener() {
            @Override
            public void onFocusChange(View v, boolean hasFocus) {
                if(hasFocus) {
                    if(activity instanceof ConfigureRules) {
                        ConfigureRules configureRules = ((ConfigureRules)activity);
                        if(configureRules.getAlertDialog() != null) {
                            ((ConfigureRules)activity).getAlertDialog().getWindow().clearFlags(
                                    WindowManager.LayoutParams.FLAG_NOT_FOCUSABLE|
                                    WindowManager.LayoutParams.FLAG_ALT_FOCUSABLE_IM);
                        }
                    }
                }
            }
        });

        editText.setOnEditorActionListener(new TextView.OnEditorActionListener() {
            @Override
            public boolean onEditorAction(TextView v, int actionId, KeyEvent event) {
                if (actionId == EditorInfo.IME_ACTION_DONE || actionId == EditorInfo.IME_ACTION_NEXT) {
                    ElementConnector.saveValueAsParameter(v.getId(), ((EditText) v).getText().toString().trim(), v.getTag(), null);
                }
                return false;
            }
        });
        editText.requestFocus();
        return editText;
    }

    public static Spinner createRFIDSpinner(Activity activity, String startValue){
        final Spinner spinner = new Spinner(activity);
        ArrayAdapter<RFIDGroup> adapter = new ArrayAdapter<RFIDGroup>(activity, android.R.layout.simple_list_item_1);
        spinner.setAdapter(adapter);
        List<RFIDGroup> rfidGroups = Model.getInstance().getActiveBoard().getHardwareSetup().getRFIDGroups();
        RFIDGroup rfidGroupStartValue = null;
        if (startValue != null && !startValue.equalsIgnoreCase("")) {
            rfidGroupStartValue = Model.getInstance().getActiveBoard().getHardwareSetup().getRFIDGroupByName(startValue);
        }
        for(RFIDGroup rfidGroup : rfidGroups) {
            adapter.add(rfidGroup);
            if (rfidGroupStartValue != null && rfidGroup == rfidGroupStartValue) {
                spinner.setSelection(adapter.getPosition(rfidGroup));
            }
        }

        adapter.notifyDataSetChanged();

        spinner.setOnItemSelectedListener(new AdapterView.OnItemSelectedListener() {
            @Override
            public void onItemSelected(AdapterView<?> adapterView, View view, int i, long l) {
                ElementConnector.saveValueAsParameter(spinner.getId(), ((RFIDGroup) spinner.getSelectedItem()).getRFIDTags(), spinner.getTag(), ((RFIDGroup) spinner.getSelectedItem()).getName());
            }

            @Override
            public void onNothingSelected(AdapterView<?> adapterView) {

            }
        });
        return spinner;
    }
    public static String getSelectedItemFromSpinner(Spinner spinner){
        if(spinner.getSelectedItem() != null){
            return ((RFIDGroup) spinner.getSelectedItem()).getRFIDTags();
        } else if(spinner.getCount() > 0 ){
            return ((RFIDGroup) spinner.getItemAtPosition(0)).getRFIDTags();
        }
        return "";
    }

}