/*
 * Licensed to UbiCollab.org under one or more contributor
 * license agreements.  See the NOTICE file distributed
 * with this work for additional information regarding
 * copyright ownership. UbiCollab.org licenses this file
 * to you under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package no.ntnu.android.toolkit.graphic;

import android.app.Activity;
import android.app.ProgressDialog;
import android.content.DialogInterface;
import android.os.Handler;
import android.view.View;
import no.ntnu.android.toolkit.model.Board;
import no.ntnu.android.toolkit.model.Model;
import no.ntnu.android.toolkit.uploader.RulesInstaller;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class InstallationProgressDialogHandler {
    private int totalStepInEachTask;
    private RulesInstaller.InstallType type;
    private List<String> messages;
    private ProgressDialog progressDialog;
    private Activity activity;
    private int delayBeforeDismiss = 5000;
    private Map<EventType, Integer> eventCountMap;
    private static int showNumberOflines = 4;


    public enum EventType {
        GENERATING,
        GETTING_IP,
        GETTING_IP_DONE,
        GETTING_IP_FAIL_WRONG_SETTINGS,
        GETTING_IP_FAIL_NOT_SUPPORTED,
        NO_RULES,
        COMPILE_START,
        DEVICE_UPLOAD_START,
        DONE,
        COMPILE_FAIL,
        DEVICE_UPLOAD_FAILED,
        DEVICE_UPLOAD_DONE,
        COMPILE_DONE,
        UPDATE_STATUS_BY_10
    }

    public InstallationProgressDialogHandler(Activity activity, RulesInstaller.InstallType type) {
        this.activity = activity;
        this.type = type;
        messages = new ArrayList<String>();
        totalStepInEachTask = 0;
        eventCountMap = new HashMap<EventType, Integer>();
    }

    private String getPostFixMessage(int currentStep) {
        return  " (" + currentStep + " / " + totalStepInEachTask + ")";
    }

    public void updateProcess(int stageValue, int currentStep) {
        int totalNew;
        if (progressDialog != null) {
            int currentValue = progressDialog.getProgress();
            if (totalStepInEachTask != 0) {
                totalNew = (currentValue + stageValue) - (stageValue - stageValue*(currentStep / totalStepInEachTask));
            } else {
                totalNew = 0;
            }
            if(totalNew <= 100 && totalNew >= 0) {
                progressDialog.setProgress(totalNew);
            } else if(totalNew > 100) {
                progressDialog.setProgress(100);
                System.err.println("Progressbar value was set to: " + totalNew + "(adjusted to 100)");
            }
        }
    }

    private String buildMessageFromMessages() {
        int start = (messages.size()-showNumberOflines) >= 0 ? (messages.size()-showNumberOflines) : 0;
        StringBuilder builder = new StringBuilder();
        for(int i = start; i < messages.size(); i++) {
            if(i != messages.size()-1) {
                builder.append(messages.get(i) + "\n");
            } else {
                builder.append(messages.get(i));
            }
        }
        return builder.toString();
    }

    private void addDone() {
        String last = messages.get(messages.size()-1);
        messages.remove(messages.size()-1);
        messages.add(last + " DONE");
        progressDialog.setMessage(buildMessageFromMessages());
    }

    private void addToPrevousMessage(String msg, boolean postfix, EventType eventType) {
        if(progressDialog != null) {
            String currentMessage;
            String previousMsg = buildMessageFromMessages();
            if(postfix) {
                currentMessage = msg + getPostFixMessage(eventCountMap.get(eventType));
            } else {
                currentMessage = msg;
            }
            progressDialog.setMessage(previousMsg + "\n" + currentMessage);
            messages.add(currentMessage);
        } else {
            throw new IllegalStateException("No progress dialog active");
        }
    }

    public void showInstallationProgressDialog() {
        progressDialog = new ProgressDialog(activity);
        progressDialog.setTitle("Installing..");
        if (type == RulesInstaller.InstallType.Single) {
            totalStepInEachTask = 1;
        } else if(type == RulesInstaller.InstallType.Interconnected ) {
            totalStepInEachTask = Model.getInstance().getBoardsWithRules().size();
        } else if(type == RulesInstaller.InstallType.MultipleEqual) {
            for(Board board : Model.getInstance().getBoards()) {
                if(board.getHardwareSetup() == Model.getInstance().getActiveBoard().getHardwareSetup()) {
                    totalStepInEachTask++;
                }
            }
        }

        progressDialog.setProgressStyle(ProgressDialog.STYLE_HORIZONTAL);
        progressDialog.setCancelable(false);
        progressDialog.setButton(DialogInterface.BUTTON_NEUTRAL, "Ok", new DialogInterface.OnClickListener() {
            @Override
            public void onClick(DialogInterface dialog, int which) {
                dialog.dismiss();
            }
        });
        progressDialog.setProgress(0);
        addToPrevousMessage("Preparing..", false, null);
        progressDialog.show();
        progressDialog.getButton(DialogInterface.BUTTON_NEUTRAL).setVisibility(View.INVISIBLE);
    }

    public void installationFeedbackChangeEvent(EventType eventType) {
        int next;
        if(eventCountMap.get(eventType) != null) {
            next = eventCountMap.get(eventType) + 1;
        } else {
            next = 1;
        }
        eventCountMap.put(eventType, next);

        switch (eventType) {
            case GENERATING:
                addToPrevousMessage("Process started", false, eventType);
                updateProcess(10, totalStepInEachTask);
                break;

            case GETTING_IP:
                addToPrevousMessage("Getting IP", true, eventType);
                updateProcess(10, eventCountMap.get(eventType));
                break;

            case GETTING_IP_DONE:
                addDone();
                updateProcess(10, eventCountMap.get(eventType));
                break;

            case GETTING_IP_FAIL_WRONG_SETTINGS:
                addToPrevousMessage("Error: Could not connect to the local WiFi network", false, eventType);
                progressDialog.getButton(DialogInterface.BUTTON_NEUTRAL).setVisibility(View.VISIBLE);
                break;

            case GETTING_IP_FAIL_NOT_SUPPORTED:
                addToPrevousMessage("Error: The app you are trying to install require a board with WiFi device. I" +
                        "f the board you are trying to install on have a wifi device, please install an app that does " +
                        "not require WiFi to fix this problem.", false, eventType);
                progressDialog.getButton(DialogInterface.BUTTON_NEUTRAL).setVisibility(View.VISIBLE);
                break;

            case NO_RULES:
                addToPrevousMessage("Error: Nothing to install", false, eventType);
                progressDialog.getButton(DialogInterface.BUTTON_NEUTRAL).setVisibility(View.VISIBLE);
                break;

            case COMPILE_START:
                addToPrevousMessage("Compiling", type != RulesInstaller.InstallType.MultipleEqual, eventType);
                updateProcess(10, eventCountMap.get(eventType));
                break;

            case COMPILE_FAIL:
                addToPrevousMessage("Error: Compilation failed", true, eventType);
                progressDialog.getButton(DialogInterface.BUTTON_NEUTRAL).setVisibility(View.VISIBLE);
                break;

            case COMPILE_DONE:
                addDone();
                updateProcess(10, eventCountMap.get(eventType));
                break;

            case DEVICE_UPLOAD_START:
                addToPrevousMessage("Uploading", true, eventType);
                updateProcess(10, eventCountMap.get(eventType));
                break;

            case DEVICE_UPLOAD_FAILED:
                addToPrevousMessage("Error: Upload failed", true, eventType);
                progressDialog.getButton(DialogInterface.BUTTON_NEUTRAL).setVisibility(View.VISIBLE);
                break;

            case DEVICE_UPLOAD_DONE:
                addDone();
                updateProcess(40, eventCountMap.get(eventType));
                break;

            case DONE:
                progressDialog.setProgress(100);
                addToPrevousMessage("Successfully installed", false, eventType);
                delayedDismiss(delayBeforeDismiss);
                break;

            case UPDATE_STATUS_BY_10:
                updateProcess(10, totalStepInEachTask);
                break;

            default:
                break;
        }
    }

    private void delayedDismiss(int delay) {
        new Handler().postDelayed(new Runnable() {
            public void run() {
                progressDialog.dismiss();
            }
        }, delay);
    }
}