/*
 * Licensed to UbiCollab.org under one or more contributor
 * license agreements.  See the NOTICE file distributed
 * with this work for additional information regarding
 * copyright ownership. UbiCollab.org licenses this file
 * to you under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package no.ntnu.android.toolkit.graphic.listener;

import android.graphics.Point;
import android.view.MotionEvent;
import android.view.View;
import android.view.ViewGroup;
import no.ntnu.android.toolkit.activities.ConfigureRules;
import no.ntnu.android.toolkit.activities.R;
import no.ntnu.android.toolkit.graphic.view.ConfigurationLinearLayout;
import no.ntnu.android.toolkit.graphic.view.ConnectorLinearLayout;
import no.ntnu.android.toolkit.model.Action;
import no.ntnu.android.toolkit.model.Condition;

public class ConnectedLinearLayoutOnTouchListener implements View.OnTouchListener {

    private final ConfigureRules configRules;
    private long lastActionTouchDown;
    private static final long minMillisTimeBetweenTouch = 300;

    public ConnectedLinearLayoutOnTouchListener(ConfigureRules configureRules) {
        this.configRules = configureRules;
    }


    @Override
    public boolean onTouch(View view, MotionEvent motionEvent) {
        if (motionEvent.getAction() == MotionEvent.ACTION_DOWN) {
            if(lastActionTouchDown != 0 && motionEvent.getEventTime() < (lastActionTouchDown + minMillisTimeBetweenTouch)) {
                return false;
            }
            configRules.disableRuleContainerScroll();
            lastActionTouchDown = motionEvent.getEventTime();
        } else if (motionEvent.getAction() == MotionEvent.ACTION_UP) {
            view.setVisibility(View.VISIBLE);
            configRules.enableRuleContainerScroll();
        }
        if (view instanceof ConfigurationLinearLayout) {
            if (((ConfigurationLinearLayout)view).getConfiguration() instanceof Condition) {
                return handleConditionLinearLayoutOnTouch((ConfigurationLinearLayout) view, motionEvent);
            } else if (((ConfigurationLinearLayout)view).getConfiguration() instanceof Action){
                return handleActionLinearLayoutOnTouch((ConfigurationLinearLayout) view, motionEvent);
            }
        } else if (view instanceof ConnectorLinearLayout) {
            return handleConnectorLinearLayoutOnTouch((ConnectorLinearLayout) view, motionEvent);
        }
        return false;
    }

    private boolean handleConditionLinearLayoutOnTouch(ConfigurationLinearLayout conditionLL, MotionEvent motionEvent) {
        if (motionEvent.getAction() == MotionEvent.ACTION_DOWN){
            if (configRules.getMode() == ConfigureRules.Mode.DRAG) {
                conditionLL.setVisibility(View.INVISIBLE);
                conditionLL.startDrag(null, new View.DragShadowBuilder(conditionLL), conditionLL, 0);
            } else {
                configRules.drawArrowBetweenConnectedLLAndPoint(conditionLL, getPointOfMotionEvent(motionEvent, conditionLL));
            }
            return true;
        }
        else if (motionEvent.getAction() == MotionEvent.ACTION_MOVE) {
            if (conditionLL.getTouchArrow() != null) {
                conditionLL.getTouchArrow().setTargetPoint(getPointOfMotionEvent(motionEvent, conditionLL));
                conditionLL.getTouchArrow().update();
            }
        }

        else if (motionEvent.getAction() == MotionEvent.ACTION_UP) {
            ViewGroup background = (ViewGroup) configRules.findViewById(R.id.rule_container);
            background.removeView(conditionLL.getTouchArrow());
            conditionLL.setTouchArrow(null);

            if (isPointInView(getPointOfMotionEvent(motionEvent, conditionLL), conditionLL) &&
                    configRules.getMode() != ConfigureRules.Mode.DRAG) {
                configRules.configureParameters(conditionLL);
                return false;
            }

            for (int i = 0; i < background.getChildCount(); i++) {
                View targetIV = background.getChildAt(i);
                if (targetIV instanceof ConfigurationLinearLayout &&
                        ((ConfigurationLinearLayout)targetIV).getConfiguration() instanceof Action &&
                        isPointInView(getPointOfMotionEvent(motionEvent, conditionLL), targetIV)) {
                    if (!conditionLL.isConnectedToCLLThroughConnectorLL((ConfigurationLinearLayout) targetIV)){
                        configRules.initialConnection(conditionLL, (ConfigurationLinearLayout) targetIV);
                    }
                }
//                else if(targetIV instanceof ConnectorLinearLayout &&
//                        isPointInView(getPointOfMotionEvent(motionEvent, conditionLL), targetIV) &&
//                        !conditionLL.isNeighborOf((ConnectedLinearLayout) targetIV) &&
//                        !conditionLL.isConnectedTargetsNeighborsNeighborOfConnectorLL((ConnectorLinearLayout) targetIV)) {
//                   configRules.connectConditionWithConnector(conditionLL, (ConnectorLinearLayout) targetIV);
//                }
            }
        }
        return false;
    }

    private boolean handleActionLinearLayoutOnTouch(ConfigurationLinearLayout actionLL, MotionEvent motionEvent) {
        if (motionEvent.getAction() == MotionEvent.ACTION_DOWN){
            if (configRules.getMode() == ConfigureRules.Mode.DRAG) {
                actionLL.setVisibility(View.INVISIBLE);
                actionLL.startDrag(null, new View.DragShadowBuilder(actionLL), actionLL, 0);
            }
            return true;

        } else if (motionEvent.getAction() == MotionEvent.ACTION_UP) {
            if (isPointInView(getPointOfMotionEvent(motionEvent, actionLL), actionLL) &&
                    configRules.getMode() != ConfigureRules.Mode.DRAG) {
                configRules.configureParameters(actionLL);
                return false;
            }
        }
        return false;
    }

    private boolean handleConnectorLinearLayoutOnTouch(ConnectorLinearLayout connectorIV, MotionEvent motionEvent) {
        if (motionEvent.getAction() == MotionEvent.ACTION_DOWN){
            if (configRules.getMode() == ConfigureRules.Mode.DRAG) {
                connectorIV.setVisibility(View.INVISIBLE);
                connectorIV.startDrag(null, new View.DragShadowBuilder(connectorIV), connectorIV, 0);
            } else {
                configRules.drawArrowBetweenConnectedLLAndPoint(connectorIV, getPointOfMotionEvent(motionEvent, connectorIV));
            }
        }
        else if (motionEvent.getAction() == MotionEvent.ACTION_MOVE) {
            if (connectorIV.getTouchArrow() != null) {
                connectorIV.getTouchArrow().setTargetPoint(getPointOfMotionEvent(motionEvent, connectorIV));
                connectorIV.getTouchArrow().update();
            }
        }

        else if (motionEvent.getAction() == MotionEvent.ACTION_UP) {
            ViewGroup background = (ViewGroup) configRules.findViewById(R.id.rule_container);
            background.removeView(connectorIV.getTouchArrow());
            connectorIV.setTouchArrow(null);

            for (int i = 0; i < background.getChildCount(); i++) {
                View targetIV = background.getChildAt(i);
                if (targetIV instanceof ConfigurationLinearLayout &&
                        ((ConfigurationLinearLayout)targetIV).getConfiguration() instanceof Action &&
                        isPointInView(getPointOfMotionEvent(motionEvent, connectorIV), targetIV)) {
                        configRules.connectActionWithConnector(connectorIV, (ConfigurationLinearLayout) targetIV);
                }
            }
        }
        return false;
    }



    private boolean isPointInView(final Point point, final View view) {
        float eventPosX = point.x;
        float eventPosY = point.y;
        float viewPosLeft = view.getX();
        float viewPosTop = view.getY();
        float viewPosRight = view.getX() + view.getWidth();
        float viewPosBot = view.getY() + view.getHeight();
        if (eventPosX < viewPosLeft) return false;
        if (eventPosX > viewPosRight) return false;
        if (eventPosY < viewPosTop) return false;
        if (eventPosY > viewPosBot) return false;
        return true;
    }

    private Point getPointOfMotionEvent(MotionEvent motionEvent, View triggerView) {
        return new Point((int)(motionEvent.getX() + triggerView.getX()), (int)(motionEvent.getY() + triggerView.getY()));
    }
}