/*
 * Licensed to UbiCollab.org under one or more contributor
 * license agreements.  See the NOTICE file distributed
 * with this work for additional information regarding
 * copyright ownership. UbiCollab.org licenses this file
 * to you under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package no.ntnu.android.toolkit.model;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;

public class Board implements IdInterface, Serializable {

    private final int id;
    private String name;
    private HardwareSetup hardwareSetup;
    private List<Rule> rules;
    private transient WifiSettings wifiSettings;
    private transient String bluetoothMacAddress;
    private boolean nameChangedByUser;

    public Board(HardwareSetup hardwareSetup) {
        this.hardwareSetup = hardwareSetup;
        if(hardwareSetup != null) {
            if(hardwareSetup.getName() != null) {
                if(hardwareSetup.getName().length() > 15) {
                    this.name = hardwareSetup.getName().substring(0, 15);
                } else {
                    this.name = hardwareSetup.getName();
                }
            }
        }
        this.rules = new ArrayList<Rule>();
        id = IdGenerator.getInstance().getNext(this);
        for (Device device : hardwareSetup.getDevices()) {
            if (device.getType().equalsIgnoreCase("wifi")) {
                wifiSettings = new WifiSettings();
            }
        }
    }

    public Board(HardwareSetup hardwareSetup, List rules) {
        this(hardwareSetup);
        if(rules != null) {
            this.rules = rules;
        }
    }

    public List<Rule> getRules(){
        if(rules == null) {
            this.rules = new ArrayList<Rule>();
        }
        return rules;
    }

    public boolean removeRule(Rule rule) {
        return rules.remove(rule);
    }

    public void clearRules() {
        rules = new ArrayList<Rule>();
    }

    public void addRule(Rule rule) {
        if(rules == null) {
            this.rules = new ArrayList<Rule>();
        }
        rules.add(rule);

    }

    public String toString() {
        return name;
    }

    public HardwareSetup getHardwareSetup() {
        return hardwareSetup;
    }

    public WifiSettings getWifiSettings() {
        if (wifiSettings == null) {
            wifiSettings = new WifiSettings();;
        }
        return wifiSettings;
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public Device getWifiDevice(){
        for(Device device : getHardwareSetup().getDevices()){
            if(device.getType().equalsIgnoreCase("wifi")){
                return device;
            }
        }
        return null;
    }

    @Override
    public int getId() {
        return id;
    }

    public String getBluetoothMacAddress() {
        return bluetoothMacAddress;
    }

    public void setBluetoothMacAddress(String bluetoothMacAddress) {
        this.bluetoothMacAddress = bluetoothMacAddress;
    }

    public boolean isNameChangedByUser() {
        return nameChangedByUser;
    }

    public void setNameChangedByUser(boolean nameChangedByUser) {
        this.nameChangedByUser = nameChangedByUser;
    }

    public boolean rulesRequireWifi() {
        for (Rule rule : rules) {
            for (Configuration configuration : rule.getConfigurations()) {
                if (configuration.getDevice().getType().equalsIgnoreCase("wifi")) {
                    return true;
                }
            }
        }
        return false;
    }

    public boolean rulesContainsDeviceFunction(String deviceFunctionName) {
        for (Rule rule : rules) {
            for (Configuration configuration : rule.getConfigurations()) {
                if (configuration.getDeviceFunction().getName().equalsIgnoreCase(deviceFunctionName)) {
                    return true;
                }
            }
        }
        return false;
    }
}