/*
 * Licensed to UbiCollab.org under one or more contributor
 * license agreements.  See the NOTICE file distributed
 * with this work for additional information regarding
 * copyright ownership. UbiCollab.org licenses this file
 * to you under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package no.ntnu.android.toolkit.model;

import android.app.Activity;
import no.ntnu.android.toolkit.appstore.SavedToolkitApplication;
import no.ntnu.android.toolkit.graphic.ConnectorIdGenerator;
import no.ntnu.android.toolkit.reader.BoardTypeReader;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

public class Model {

    private static Model model = null;
    private static List<Board> boards = null;
    private static WifiSettings wifiSettings;
    private static Board active;
    private static byte[] draft;

    private Model() {    }

    public static Model getInstance() {
        if (model == null) {
            model = new Model();
            boards = new ArrayList<Board>();
        }
        return model;
    }

    public Board getActiveBoard() {
        return active;
    }

    public List<Board> getBoards() {
        return new ArrayList<Board>(boards);
    }

    public List<Board> getBoardsWithRules() {
        List<Board> boardsWithRules = new ArrayList<Board>();
        for (Board board : boards) {
            if (board.getRules() != null && board.getRules().size() != 0) {
                boardsWithRules.add(board);
            }
        }
        return boardsWithRules;
    }

    public boolean addBoard(Board board) {
        if(boards != null && !boards.contains(board)) {
            boards.add(board);
            if(active == null) {
                setActive(board);
            }
            return true;
        }
        return false;
    }

    public Board createAddBoard(HardwareSetup hardwareSetup) {
        Board newBoard = new Board(hardwareSetup);
        addBoard(newBoard);
        return newBoard;
    }

    public boolean setActive(Board board) {
        if(boards != null && board != null) {
            if(!boards.contains(board)) {
                addBoard(board);
            }
            active = board;
            return true;
        }
        return false;
    }

    public void clearAllRules() {
        for(Board board : boards) {
            board.clearRules();
        }
    }

    public void clearAllBoards() {
        boards = new ArrayList<Board>();
        active = null;
    }

    public void setWifiSettings(WifiSettings settings) {
        wifiSettings = settings;
    }

    public WifiSettings getWifiSettings() {
        return wifiSettings;
    }

    public List<Board> getAllBoardsWithWifi() {
        List<Board> boardArrayList = new ArrayList<Board>();
        for(Board board : boards) {
            if (board.getWifiDevice() != null) {
                boardArrayList.add(board);
            }
        }
        return boardArrayList;
    }

    public boolean someRulesRequiresWifi() {
        for (Board board : Model.getInstance().getBoardsWithRules()) {
            if (board.rulesRequireWifi()) {
                return true;
            }
        }
        return false;
    }

    public boolean configurationInRules(Configuration configuration) {
        for (Board board : getBoardsWithRules()) {
            for (Rule rule : board.getRules()) {
                if (rule.getConfigurations().contains(configuration)) {
                    return true;
                }
            }
        }
        return false;
    }

    public byte[] getDraft() {
        return draft;
    }

    public void setDraft(byte[] application) {
        draft = application;
    }

    public static void updateModel(SavedToolkitApplication savedToolkitApplication, Activity activity) {
        if(savedToolkitApplication != null) {
            Model.getInstance().clearAllBoards();
            IdGenerator.getInstance().flush();
            ConnectorIdGenerator.getInstance().flush();
            for(Board board : savedToolkitApplication.getBoards()) {
                Model.getInstance().addBoard(board);
                IdGenerator.getInstance().getNext(board);
                IdGenerator.getInstance().getNext(board.getHardwareSetup());
                for (Device device : board.getHardwareSetup()) {
                    IdGenerator.getInstance().getNext(device);
                }
                for (Rule rule : board.getRules()) {
                    IdGenerator.getInstance().getNext(rule);
                    for (Configuration configuration : rule.getConfigurations()) {
                        IdGenerator.getInstance().getNext(configuration);
                    }
                }
                if (savedToolkitApplication.isSaveBTMac()) {
                    HashMap<Board, String> boardMac = savedToolkitApplication.getBoardBTMacAddresses();
                    if (boardMac != null && boardMac.keySet().size() > 0) {
                        board.setBluetoothMacAddress(boardMac.get(board));
                    }
                }
            }
        }
        //Add device functions to boards in model
        BoardTypeReader.addDeviceFunctionsToBoards(activity);
    }
}