/*
 * Licensed to UbiCollab.org under one or more contributor
 * license agreements.  See the NOTICE file distributed
 * with this work for additional information regarding
 * copyright ownership. UbiCollab.org licenses this file
 * to you under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package no.ntnu.android.toolkit.uploader;

import android.app.Activity;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothSocket;
import android.content.Intent;
import android.os.Build;
import android.os.Handler;
import android.os.Message;
import no.ntnu.android.toolkit.uploader.stk500_v1.Logger;
import no.ntnu.android.toolkit.uploader.stk500_v1.STK500v1;

import java.io.*;
import java.lang.reflect.Method;
import java.util.UUID;

public class ArduinoBluetoothHandler {

    private OutputStream outputStream;
    private InputStream inputStream;
    private BluetoothSocket btSocket;
    private BluetoothAdapter btAdapter;
    private Log log;
    private UUID MY_UUID;
    private long timeout = 20000;
    private boolean isReady;

    private static ArduinoBluetoothHandler arduinoBluetoothHandler;

    private Handler handler = new Handler() {
        @Override
        public void handleMessage(Message msg) {
            super.handleMessage(msg);
        }
    };

    public static ArduinoBluetoothHandler getInstance() {
        if (arduinoBluetoothHandler == null) {
            arduinoBluetoothHandler = new ArduinoBluetoothHandler();
            arduinoBluetoothHandler.isReady = true;
        }
        return arduinoBluetoothHandler;
    }

    private ArduinoBluetoothHandler() {
        this.MY_UUID = UUID.fromString("00001101-0000-1000-8000-00805F9B34FB");
        this.log = new Log(getClass().getSimpleName());
        btAdapter = BluetoothAdapter.getDefaultAdapter();
    }

    public boolean uploadToArduino(final String hexData, final String macAddress, final Activity currentActivity) {
        isReady = false;
        checkBTState(currentActivity);
        final long start = System.currentTimeMillis();
        while (!connect(macAddress)) {
            if (start + timeout < System.currentTimeMillis()) {
                isReady = true;
                return false;
            }
        }

        //Convert string to byte array
        byte[] binaryFile = new byte[hexData.length() / 2];
        for(int i=0; i < hexData.length(); i+=2)
        {
            binaryFile[i/2] = Integer.decode("0x" + hexData.substring(i, i + 2)).byteValue();
        }

        STK500v1 p = new STK500v1(outputStream, inputStream, log, binaryFile);
        if (!p.programUsingOptiboot(false, 128)) {
            disconnect();
            handler.sendEmptyMessage(0);
            return false;
        }
        disconnect();
        handler.sendEmptyMessage(0);
        return true;
    }

    public void disconnect() {
        try {
            isReady = true;
            btSocket.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public String getIpFromArduino(String macAddress, String ssid, String passphrase) {
        if (!connect(macAddress)) {
            isReady = true;
            return null;
        }
        try {
            outputStream.write(0xFD);
            outputStream.write((ssid + ":" + passphrase).getBytes());
            outputStream.flush();
        } catch (IOException e) {
            e.printStackTrace();
        }
        String ipAddress = "";
        InputStreamReader inputStreamReader = new InputStreamReader(inputStream);
        while (true) {
            try {
                int next = inputStreamReader.read();
                if (next == 65533) {
                    inputStreamReader.close();
                    break;
                } else {
                    ipAddress += (char)next;
                }
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
        disconnect();
        if (ipAddress.equals("null") || ipAddress.equals("")) {
            return null;
        }
        return ipAddress;
    }

    private boolean connect(String macAddress) {
        isReady = false;
        BluetoothDevice device = btAdapter.getRemoteDevice(macAddress);

        try {
            btSocket = createBluetoothSocket(device);
        } catch (IOException e1) {
            return false;
        }

        btAdapter.cancelDiscovery();

        try {
            btSocket.connect();
        } catch (IOException e) {
            try {
                btSocket.close();
            } catch (IOException e2) {
            }
            return false;
        }

        try {
            outputStream = btSocket.getOutputStream();
        } catch (IOException e) {
            return false;
        }
        try {
            inputStream = btSocket.getInputStream();
        } catch (IOException e) {
            return false;
        }
        return true;
    }

    private BluetoothSocket createBluetoothSocket(BluetoothDevice device) throws IOException {
        if(Build.VERSION.SDK_INT >= 10){
            try {
                final Method m = device.getClass().getMethod("createInsecureRfcommSocketToServiceRecord", new Class[] { UUID.class });
                return (BluetoothSocket) m.invoke(device, MY_UUID);
            } catch (Exception e) {
            }
        }
        return  device.createRfcommSocketToServiceRecord(MY_UUID);
    }

    private void checkBTState(Activity currentActivity) {
        if(btAdapter != null) {
            if (!btAdapter.isEnabled()) {
                Intent enableBtIntent = new Intent(BluetoothAdapter.ACTION_REQUEST_ENABLE);
                currentActivity.startActivityForResult(enableBtIntent, BluetoothAdapter.STATE_CONNECTING);
            }
        }
    }

    public boolean isReady() {
        return isReady;
    }

    private class Log implements Logger {

        private String TAG;

        public Log(String tag) {
            this.TAG = tag;
        }

        @Override
        public void makeToast(String msg) {

        }

        @Override
        public void printToConsole(String msg) {

        }

        @Override
        public void logcat(String msg, String level) {
            if (level.equals("v")) {
                android.util.Log.v(TAG, msg);
            }
            else if (level.equals("d")){
                android.util.Log.d(TAG, msg);
            }
            else if (level.equals("i")) {
                android.util.Log.i(TAG, msg);
            }
            else if (level.equals("w")) {
                android.util.Log.w(TAG, msg);
            }
            else if (level.equals("e")) {
                android.util.Log.e(TAG, msg);
            }
        }
    }
}
