/*
 * Licensed to UbiCollab.org under one or more contributor
 * license agreements.  See the NOTICE file distributed
 * with this work for additional information regarding
 * copyright ownership. UbiCollab.org licenses this file
 * to you under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package no.ntnu.android.toolkit.uploader;

import android.os.AsyncTask;
import no.ntnu.android.toolkit.activities.ConfigureRules;
import no.ntnu.android.toolkit.connection.CompileServerConnection;
import no.ntnu.android.toolkit.generator.Generator;
import no.ntnu.android.toolkit.graphic.InstallationProgressDialogHandler;
import no.ntnu.android.toolkit.model.Board;
import no.ntnu.android.toolkit.model.Model;
import no.ntnu.android.toolkit.model.Rule;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

public class RulesInstaller extends AsyncTask<Void, Void, Void> {
    private ArduinoBluetoothHandler uploader;
    private ConfigureRules configureRules;
    private InstallType installType;
    private InstallationProgressDialogHandler barHandler;
    private boolean wifiSettingsWrong;
    private boolean ipLookupNotSupported;

    public enum InstallType {
        Single,
        MultipleEqual,
        Interconnected,
    }

    public RulesInstaller(ConfigureRules configureRules, InstallType installType,
                          InstallationProgressDialogHandler installationProgressDialogHandler){
        super();
        this.uploader = ArduinoBluetoothHandler.getInstance();
        this.configureRules = configureRules;
        this.installType = installType;
        this.barHandler = installationProgressDialogHandler;
    }

    private Map<String, String> getSingleCppCode() {
        Map<String, String> macAdressToCppCode = new HashMap<String, String>();
        if (Model.getInstance().getActiveBoard().getRules() != null && Model.getInstance().getActiveBoard().getRules().size() != 0) {
            if (Model.getInstance().getActiveBoard().rulesRequireWifi()) {
                Model.getInstance().getActiveBoard().getWifiSettings().setIpAddress(getIpAddress(Model.getInstance().getActiveBoard()));
            } else {
                installationBarChangeEvent(InstallationProgressDialogHandler.EventType.UPDATE_STATUS_BY_10);
            }
            Generator generator = new Generator(new ArrayList(Model.getInstance().getActiveBoard().getRules()), configureRules.getAssets());
            macAdressToCppCode.put(Model.getInstance().getActiveBoard().getBluetoothMacAddress(), generator.generateCppCode());
        }
        return macAdressToCppCode;
    }

    private Map<String, String> getMultipleEqualCppCode() {
        Map<String, String> macAdressToCppCode = new HashMap<String, String>();
        if (Model.getInstance().getActiveBoard().getRules() != null && Model.getInstance().getActiveBoard().getRules().size() != 0) {
            for (Board board : Model.getInstance().getBoards()) {
                if (board.getHardwareSetup().getName().equals(Model.getInstance().getActiveBoard().getHardwareSetup().getName())) {
                    if (Model.getInstance().getActiveBoard().rulesRequireWifi()) {
                        board.getWifiSettings().setIpAddress(getIpAddress(board));
                        if (wifiSettingsWrong || ipLookupNotSupported) {
                            break;
                        }
                    } else {
                        installationBarChangeEvent(InstallationProgressDialogHandler.EventType.UPDATE_STATUS_BY_10);
                    }
                }
            }
            for(Board board : Model.getInstance().getBoards()) {
                if(board.getHardwareSetup() == Model.getInstance().getActiveBoard().getHardwareSetup()) {
                    Generator generator = new Generator((ArrayList<Rule>) Model.getInstance().getActiveBoard().getRules(), configureRules.getAssets());
                    macAdressToCppCode.put(board.getBluetoothMacAddress(), generator.generateCppCode());
                }
            }
        }
        return macAdressToCppCode;
    }

    private Map<String, String> getInterconnectCppCode() {
        Map<String, String> macAdressToCppCode = new HashMap<String, String>();
        for (Board board : Model.getInstance().getBoardsWithRules()) {
            if (board.rulesRequireWifi()) {
                board.getWifiSettings().setIpAddress(getIpAddress(board));
                if (wifiSettingsWrong || ipLookupNotSupported) {
                    break;
                }
            } else {
                installationBarChangeEvent(InstallationProgressDialogHandler.EventType.UPDATE_STATUS_BY_10);
            }
        }
        for(Board board : Model.getInstance().getBoardsWithRules()) {
            Generator generator = new Generator(new ArrayList(board.getRules()), configureRules.getAssets());
            macAdressToCppCode.put(board.getBluetoothMacAddress(), generator.generateCppCode());

        }
        return macAdressToCppCode;
    }

    @Override
    protected Void doInBackground(Void... params) {
        installationBarChangeEvent(InstallationProgressDialogHandler.EventType.GENERATING);
        boolean compiled = false;
        Map<String, String> macAdressToCppCode = null;
        if(installType == InstallType.Single) {
            macAdressToCppCode = getSingleCppCode();
        } else if(installType == InstallType.MultipleEqual) {
            macAdressToCppCode = getMultipleEqualCppCode();
        } else if(installType == InstallType.Interconnected) {
            macAdressToCppCode = getInterconnectCppCode();
        }

        if (wifiSettingsWrong) {
            installationBarChangeEvent(InstallationProgressDialogHandler.EventType.GETTING_IP_FAIL_WRONG_SETTINGS);
            wifiSettingsWrong = false;
            return null;
        }
        if (ipLookupNotSupported) {
            installationBarChangeEvent(InstallationProgressDialogHandler.EventType.GETTING_IP_FAIL_NOT_SUPPORTED);
            ipLookupNotSupported = false;
            return null;
        }

        if(macAdressToCppCode.keySet().size() == 0) {
            installationBarChangeEvent(InstallationProgressDialogHandler.EventType.NO_RULES);
            return null;
        }

        CompileServerConnection serverConnection = new CompileServerConnection();
        Iterator it = macAdressToCppCode.entrySet().iterator();
        String hex = "";
        while (it.hasNext()) {
            Map.Entry entry = (Map.Entry)it.next();
            String mac = (String)entry.getKey();
            String cpp = (String)entry.getValue();
            if(!compiled || installType == InstallType.Interconnected) {
                installationBarChangeEvent(InstallationProgressDialogHandler.EventType.COMPILE_START);
                hex = serverConnection.compile(cpp);
                if (hex == null || hex.equals("") || hex.equalsIgnoreCase("fail")) {
                    installationBarChangeEvent(InstallationProgressDialogHandler.EventType.COMPILE_FAIL);
                    return null;
                }
                installationBarChangeEvent(InstallationProgressDialogHandler.EventType.COMPILE_DONE);
                compiled = true;
            }
            if (uploader.isReady()) {
                installationBarChangeEvent(InstallationProgressDialogHandler.EventType.DEVICE_UPLOAD_START);
                if (!upload(hex, mac)) {
                    installationBarChangeEvent(InstallationProgressDialogHandler.EventType.DEVICE_UPLOAD_FAILED);
                    return null;
                } else {
                    installationBarChangeEvent(InstallationProgressDialogHandler.EventType.DEVICE_UPLOAD_DONE);
                }
            }
        }
        installationBarChangeEvent(InstallationProgressDialogHandler.EventType.DONE);
        return null;
    }

    private void installationBarChangeEvent(final InstallationProgressDialogHandler.EventType type) {
        configureRules.runOnUiThread(new Runnable() {
            @Override
            public void run() {
                barHandler.installationFeedbackChangeEvent(type);
            }
        });
    }
    private boolean upload(String hex, String  macAddress) {
        ArduinoBluetoothHandler uploader = ArduinoBluetoothHandler.getInstance();
        return uploader.uploadToArduino(hex, macAddress, configureRules);
    }


    private String getIpAddress(Board board) {
        installationBarChangeEvent(InstallationProgressDialogHandler.EventType.GETTING_IP);
        ArduinoBluetoothHandler bluetoothConnector = ArduinoBluetoothHandler.getInstance();
        String ipAddress = null;
        if (bluetoothConnector.isReady()) {
            ipAddress = bluetoothConnector.getIpFromArduino(board.getBluetoothMacAddress(), board.getWifiSettings().getSsid(),
                    board.getWifiSettings().getPassword());
            if (ipAddress == null) {
                wifiSettingsWrong = true;
            } else if (ipAddress.equalsIgnoreCase("not_supported")) {
                ipLookupNotSupported = true;
            } else {
                installationBarChangeEvent(InstallationProgressDialogHandler.EventType.GETTING_IP_DONE);
            }

        }
        return ipAddress;
    }
}